# -*- coding: utf-8 -*-
"""
Created on Fri Nov  8 15:37:53 2019

@author: Angshuman Deb
"""

from .utility import Utility
from .multivariate_nataf import multivariate_nataf
from .mixture import mixture
import numpy as np
import os
from scipy.linalg import block_diag
import scipy.io as sio
import matplotlib.pyplot as plt
from collections.abc import Sequence
import matplotlib.tri as tri
import shutil
from copy import deepcopy


class Structure:

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, name, location_info, model_files_path, model_work_dir_path, model_params, structural_analysis_platform):
        self.name = name
        self.location_info = location_info
        self.model_files_path = Utility.get_path(model_files_path)
        self.model_work_dir_path = Utility.get_path(model_work_dir_path)
        self.model_params = model_params
        self.structural_analysis_platform = structural_analysis_platform

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def get_all_dir_levels():
        return ['Analysis_Case', 'Design_Num', 'Model_Option_Num', 'Model_Realization_Num',
                'Hazard_Level_Num', 'Ground_Motion_Rec_Num']

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_dir_level_index(dir_level_to_seek):
        all_dir_levels = Structure.get_all_dir_levels()
        return all_dir_levels.index(dir_level_to_seek)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_analysis_cases(tag):
        to_return = list()
        if tag == 'Deterministic_FE_Model':
            to_return.extend(['100'])
        if tag == 'Incl_FE_Model_Param_Uncertainty':
            to_return.extend(['200'])
        if tag == 'Deterministic_FE_Model_Incl_Model_Form_Error':
            to_return.extend(['300'])
        if tag == 'Incl_Prob_Dist_Param_Estimation_Uncertainty':
            to_return.extend(['400'])
        if tag == 'All':
            to_return.extend(['100', '200', '300', '400'])

        return to_return

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_analysis_type(case):
        if case == '100':
            return 'Deterministic_FE_Model'
        if case == '200':
            return 'Incl_FE_Model_Param_Uncertainty'
        if case == '300':
            return 'Deterministic_FE_Model_Incl_Model_Form_Error'
        if case == '400':
            return 'Incl_Prob_Dist_Param_Estimation_Uncertainty'

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_modified_for_which(for_which, dir_level_to_seek):
        return for_which[0: Structure.get_dir_level_index(dir_level_to_seek) + 1]

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def get_design_num_list(self, values):
        if isinstance(values, Sequence) and not isinstance(values, str):
            d = self.model_params['primary_design_params']['value_list_dict']
            design_num_list = [key for key in d.keys() for value in values if d[key] == value]
            return design_num_list
        elif isinstance(values, str) and values == 'all':
            return list(self.model_params['primary_design_params']['value_list_dict'].keys())

    # ------------------------------------------------------------------------------------------------------------------

    def get_design_pts(self, design_num_list):
        d = self.model_params['primary_design_params']['value_list_dict']
        if isinstance(design_num_list, Sequence) and not isinstance(design_num_list, str):
            return [d[num] for num in design_num_list]
        elif isinstance(design_num_list, str) and design_num_list == 'all':
            return [d[key] for key in d.keys()]

    def add_design_point(self, value, gp=False):
        if value is None:
            return
        d = self.model_params['primary_design_params']['value_list_dict']
        if value in d.values():
            return
        max_design_num = np.max(np.array(list(d.keys())).astype(int))
        d[f'{max_design_num + 1}'] = value
        d0 = self.model_params['primary_design_params']['design_point_qualifier_dict']
        if gp:
            d0[f'{max_design_num + 1}'] = ['gp']
        else:
            d0[f'{max_design_num + 1}'] = ['non-gp']

    def set_site_hazard_info(self, im):
        if im.gmm.NAME == 'Boore and Atkinson (2008)':
            site_hazard_info_dir_path = Utility.get_path(self.model_files_path, 'Site_Hazard')

            # Get scenario data
            scenario_data = Utility.read_numpy_array_from_txt_file(
                Utility.get_path(site_hazard_info_dir_path, 'deagg_data_small_sa.txt'), skiprows='find')

            # Eliminate 0 contrib scenarios
            scenario_data = scenario_data[scenario_data[:, -1] != 0, :]

            # Eliminate dist > 300 km scenarios
            scenario_data = scenario_data[scenario_data[:, 0] <= 300, :]

            # scenario_data = [dist mag contrib]
            scenario_data = np.array([scenario_data[:, 0], scenario_data[:, 1], scenario_data[:, -1]]).T

            with open(Utility.get_path(site_hazard_info_dir_path, 'deagg_summary_small_sa.txt'), 'r') as fid:
                curr_line = fid.readline()

            hazard_prob_exceedance = float(curr_line.split(')')[0].split()[-1])
            # TODO: include T in formula
            hazard_mar = -np.log(1 - hazard_prob_exceedance)

            scenario_data[:, -1] = (scenario_data[:, -1] / 100) * hazard_mar

            param_names = im.get_gmm_param_names()

            site_hazard_info = dict()
            site_hazard_info['scenario_rate'] = scenario_data[:, -1]
            site_hazard_info['scenario_data'] = dict()
            site_hazard_info['scenario_data'][param_names[0]] = scenario_data[:, 1]
            site_hazard_info['scenario_data'][param_names[1]] = scenario_data[:, 0]
            site_hazard_info['scenario_data'][param_names[2]] = np.array([self.location_info['v_s30']] * scenario_data.shape[0])
            site_hazard_info['scenario_data'][param_names[3]] = np.array([self.location_info['mechanism']] * scenario_data.shape[0])
            site_hazard_info['scenario_data']['region'] = np.array([self.location_info['region']] * scenario_data.shape[0])

            Utility.pickle_dump_dict(Utility.get_path(site_hazard_info_dir_path, 'site_hazard_info.pickle'),
                                     site_hazard_info)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def get_site_hazard_info(self):
        site_hazard_info_file_path = Utility.get_path(self.model_files_path, 'Site_Hazard',
                                                      'site_hazard_info.pickle')
        return Utility.pickle_load_dict(site_hazard_info_file_path)

    # ------------------------------------------------------------------------------------------------------------------

    def generate_random_model_param_vals(self, analysis_case, design_num, **kwargs):
        rng_seed_recvd = kwargs.get('rng_seed', None)
        save = kwargs.get('save', True)
        sample_size = kwargs.get('sample_size', None)
        sampling_method = kwargs.get('sampling_method', None)

        model_params = self.model_params
        model_work_dir_path = self.model_work_dir_path
        dir_category = 'Random_Model_Param_Vals'

        if sample_size is None:
            sample_size = model_params['random_model_params']['sample_size']
        if sampling_method is None:
            sampling_method = model_params['random_model_params']['sampling_method']

        prob_dist_list = self.model_params['random_model_params']['prob_dist_list']
        corr_matrix = self.model_params['random_model_params']['corr_matrix']
        dist_params_sample_size = model_params['random_model_params']['dist_params_sample_size']
        estimation_sample_size = model_params['random_model_params']['estimation_sample_size']
        diag_blocks = model_params['random_model_params']['diag_blocks']

        if analysis_case in Structure.get_analysis_cases('Deterministic_FE_Model'):
            expected_value_list = self.get_flattened_rv_expected_value_list()
            random_model_param_vals = np.array(expected_value_list)[np.newaxis, :, np.newaxis]

        elif analysis_case in Structure.get_analysis_cases('Incl_FE_Model_Param_Uncertainty'):
            m0 = multivariate_nataf(prob_dist_list, corr_matrix)
            m = multivariate_nataf(self.get_flattened_rv_prob_dist_list(),
                                   self.get_corr_matrix_full(),
                                   corr_matrix_z=self.get_corr_matrix_full(corr_matrix=m0.corr_matrix_z))
            random_model_param_vals = m.rvs(size=sample_size,
                                            random_state=rng_seed_recvd, method=sampling_method)[:, :, np.newaxis]

        elif analysis_case in Structure.get_analysis_cases('Incl_Prob_Dist_Param_Estimation_Uncertainty'):
            dist_param_vals_list = list()
            corr_matrix_realizations = np.array([]).reshape((0, 0, dist_params_sample_size))
            prob_dists = np.array(prob_dist_list)
            itr_group = 0
            for diag_block in diag_blocks:
                prob_dist_block = prob_dists[np.array(diag_block)]
                corr_matrix_block = Utility.get_indexed_matrix(corr_matrix, diag_block)
                if rng_seed_recvd is None:
                    rng_seed = None
                else:
                    rng_seed = Utility.get_num([rng_seed_recvd, itr_group])
                dist_param_vals_list_temp, corr_matrix_realizations_temp = Utility.generate_random_dist_param_vals(
                    prob_dist_block,
                    corr_matrix_block,
                    estimation_sample_size,
                    dist_params_sample_size,
                    rng_seed=rng_seed
                )
                itr_group += 1
                dist_param_vals_list.extend(dist_param_vals_list_temp)
                corr_matrix_realizations = Utility.block_diag_3d(corr_matrix_realizations,
                                                                 corr_matrix_realizations_temp)
            itr_group = 0
            iter_model_option = 0
            prob_dist_list_temp = list()
            for iter_dist in range(len(prob_dist_list)):
                prob_dist_list_temp.append(
                    mixture(dist_param_vals_list[iter_dist], each=prob_dist_list[iter_dist].dist))
            m0 = multivariate_nataf(prob_dist_list_temp, corr_matrix_realizations)
            m = multivariate_nataf(self.get_flattened_rv_prob_dist_list(prob_dist_list=prob_dist_list_temp),
                                   self.get_corr_matrix_full(corr_matrix=m0.corr_matrix),
                                   corr_matrix_z=self.get_corr_matrix_full(corr_matrix=m0.corr_matrix_z))
            if rng_seed_recvd is None:
                rng_seed = None
            else:
                rng_seed = Utility.get_num([rng_seed_recvd, itr_group, iter_model_option])
            random_model_param_vals = m.rvs(size=sample_size, random_state=rng_seed,
                                            method=sampling_method)[:, :, np.newaxis]

        elif analysis_case in Structure.get_analysis_cases('Deterministic_FE_Model_Incl_Model_Form_Error'):
            num_model_options = self.get_num_model_options()
            expected_value_list = self.get_flattened_rv_expected_value_list()
            random_model_param_vals = np.array(expected_value_list)[np.newaxis, :, np.newaxis].repeat(
                num_model_options, axis=2)
        else:
            return

        if save:
            # Save random_model_param_vals
            file_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category,
                                             analysis_case, design_num)
            if not os.path.isdir(file_dir_path):
                os.makedirs(file_dir_path)
            file_path = Utility.get_path(file_dir_path, 'random_model_param_vals.pickle')
            Utility.pickle_dump_dict(file_path, {'random_model_param_vals': random_model_param_vals})
        return random_model_param_vals

    # ------------------------------------------------------------------------------------------------------------------

    def get_random_model_param_vals(self, for_which):
        model_work_dir_path = self.model_work_dir_path
        dir_category = 'Random_Model_Param_Vals'

        dir_level_to_seek = 'Design_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        work_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which))

        file_path = Utility.get_path(work_dir_path, 'random_model_param_vals.pickle')

        return Utility.pickle_load_dict(file_path)['random_model_param_vals']

    # ------------------------------------------------------------------------------------------------------------------

    def get_num_model_options(self):
        model_attributes = self.model_params['model_attributes']
        prod = 1
        for item in list(model_attributes.values()):
            prod *= len(item)
        return prod

    # ------------------------------------------------------------------------------------------------------------------

    def remove_from_work_dir(self, what, **kwargs):

        # what = 'all', 'startswith <tag>', 'contains <tag>', [list of thing(s)]

        local_bash_path = kwargs.get('local_bash_path', '')
        comp_env = kwargs.get('comp_env', 'local')

        if type(what) == str and what == 'all':
            stuff_to_remove = [f'*']

        elif type(what) == str and what.startswith('startswith '):
            tag = what.split()[1]
            stuff_to_remove = [f'{tag}*']

        elif type(what) == str and what.startswith('contains '):
            tag = what.split()[1]
            stuff_to_remove = [f'*{tag}*']

        elif type(what) == list:
            stuff_to_remove = what

        else:
            return

        find_string = ' -o '.join([f'-name "{stuff}"' for stuff in stuff_to_remove])

        model_work_dir_path = self.model_work_dir_path
        work_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir')

        exec_commands = [f'find "{work_dir_path}" \\( {find_string} \\) -exec echo -n \'"{{}}" \' \\;'
                         f' | xargs -n 50 -P 1 rm -rf']
        Utility.shell_exec('REMOVE', '', exec_commands, comp_env=comp_env, local_bash_path=local_bash_path)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def tar_work_dir(self, **kwargs):

        local_bash_path = kwargs.get('local_bash_path', '')
        exclude_list = kwargs.get('exclude_list', [])
        include_list = kwargs.get('include_list', [])
        if len(exclude_list) != 0 and len(include_list) != 0:
            raise ValueError("Cannot specify both exclude and include lists!")
        comp_env = kwargs.get('comp_env', 'local')
        new_file = kwargs.get('new_file', False)
        flags = kwargs.get('flags', ['c', 'z', 'v', 'f'])

        if new_file:
            tar_file_name = self.name + '_Work_Dir_' + Utility.get_curr_time_stamp() + '.tar'
        else:
            tar_file_name = self.name + '_Work_Dir.tar'

        tar_string = f"tar -{''.join(flags)} {tar_file_name}"

        exclude_string = ''
        for item in exclude_list:
            exclude_string += f' --exclude=\'{item}\''

        tar_string += exclude_string
        if len(include_list) == 0:
            tar_string += ' Work_Dir'
        else:
            for item in include_list:
                tar_string += f' Work_Dir/{item}'

        exec_commands = [tar_string]
        Utility.shell_exec('TAR', self.model_work_dir_path, exec_commands,
                           comp_env=comp_env, local_bash_path=local_bash_path)

        return tar_file_name

    # ------------------------------------------------------------------------------------------------------------------

    def get_material_hysteresis(self, for_which, mat_tag, input_data, num_incr):
        input_data = ' '.join([str(float(data)) for data in input_data])
        model_work_dir_path = self.model_work_dir_path
        dir_category = 'Prelim_Analysis_Results'
        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        mat_data_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, *for_which, self.structural_analysis_platform.model_info_dir_name)
        return self.structural_analysis_platform.perform_mat_test(mat_data_dir_path, mat_tag, input_data, num_incr)

    # ------------------------------------------------------------------------------------------------------------------

    def get_periods(self, for_which, modes):
        model_work_dir_path = self.model_work_dir_path
        dir_category = 'Prelim_Analysis_Results'

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        work_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which))
        periods = Utility.read_numpy_array_from_txt_file(
            Utility.get_path(work_dir_path, self.structural_analysis_platform.model_info_dir_name, 'periods.txt'),
            skiprows='find').flatten()

        if isinstance(modes, Sequence):
            mode_indices = (np.array(modes) - 1).astype(int)
        else:
            mode_indices = int(modes - 1)
        return periods[mode_indices]

    def get_period_range_extremes(self, for_which, define_range):
        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        period_extremes = list()
        for itr in define_range:
            itr_split = itr.split('_')
            if len(itr_split) == 2 and itr_split[0] == 'T' and itr_split[1].isdigit():
                period = self.get_periods(for_which, int(itr_split[1]))
            else:
                raise NotImplementedError(f"period getter not defined for {itr}")
            period_extremes.append(period)
        if len(period_extremes) == 1:
            period_extremes = [period_extremes[0], period_extremes[0]]
        return period_extremes

    def get_period_range(self, for_which, define_range, range_multiplier, num_periods, type_spacing):
        period_extremes = self.get_period_range_extremes(for_which, define_range)
        # return 1-d array
        if type_spacing == 'log':
            return np.logspace(np.log10(range_multiplier[0] * period_extremes[0]),
                               np.log10(range_multiplier[1] * period_extremes[1]), num_periods, endpoint=True)
        else:
            return np.linspace(range_multiplier[0] * period_extremes[0],
                               range_multiplier[1] * period_extremes[1], num_periods, endpoint=True)

    # ------------------------------------------------------------------------------------------------------------------

    def get_flattened_rv_name_list(self):
        rv_list = self.model_params['random_model_params']['rv_list']
        flattened_list = list()
        for itr in range(len(rv_list)):
            flattened_list.extend(rv_list[itr]['name_list'])
        return flattened_list

    # ------------------------------------------------------------------------------------------------------------------

    def get_flattened_rv_expected_value_list(self):
        rv_list = self.model_params['random_model_params']['rv_list']
        flattened_list = list()
        for itr in range(len(rv_list)):
            flattened_list.extend(rv_list[itr]['expected_value_list'] * rv_list[itr]['count'])
        return flattened_list

    # ------------------------------------------------------------------------------------------------------------------

    def get_flattened_rv_prob_dist_list(self, prob_dist_list=None):
        rv_list = self.model_params['random_model_params']['rv_list']
        flattened_list = list()
        if prob_dist_list is None:
            prob_dists = np.array(self.model_params['random_model_params']['prob_dist_list'])
        else:
            prob_dists = np.array(prob_dist_list)
        for itr in range(len(rv_list)):
            prob_dist_ind = np.array(rv_list[itr]['prob_dist_ind_list'])
            temp = list(prob_dists[prob_dist_ind])
            flattened_list.extend(temp * rv_list[itr]['count'])
        return flattened_list

    # ------------------------------------------------------------------------------------------------------------------

    def get_corr_matrix_full(self, corr_matrix=None):
        if corr_matrix is None:
            corr_matrix = self.model_params['random_model_params']['corr_matrix']
        corr_matrix_full = np.array([]).reshape(0, 0)
        rv_list = self.model_params['random_model_params']['rv_list']
        for itr in range(len(rv_list)):
            temp = Utility.get_indexed_matrix(corr_matrix, rv_list[itr]['prob_dist_ind_list'])
            corr_matrix_full = block_diag(corr_matrix_full, *([temp] * rv_list[itr]['count']))
        return corr_matrix_full

    # ------------------------------------------------------------------------------------------------------------------

    def get_rv_index(self, key, whole=False):
        name_list = self.get_flattened_rv_name_list()
        if not whole:
            index_list = [itr for itr in range(len(name_list)) if key in name_list[itr]]
        else:
            index_list = [itr for itr in range(len(name_list)) if key == name_list[itr]]
        return index_list

    # ------------------------------------------------------------------------------------------------------------------

    def get_prob_dist_list_from_flattened(self, rv_indices):
        prob_dist_list_flattened = self.get_flattened_rv_prob_dist_list()
        prob_dist_list = list()
        for ind in rv_indices:
            prob_dist_list.append(prob_dist_list_flattened[ind])
        corr_matrix_full = self.get_corr_matrix_full()
        return prob_dist_list, Utility.get_indexed_matrix(corr_matrix_full, rv_indices)

    # ------------------------------------------------------------------------------------------------------------------

    def plot_scenario_rates(self, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        cmap = kwargs.get('cmap', 'viridis')
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid = kwargs.get('minor_grid', False)
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        fig_ax = kwargs.get('fig_ax', None)

        site_hazard_info = self.get_site_hazard_info()
        rate = site_hazard_info['scenario_rate']
        mag = site_hazard_info['scenario_data']['mag']
        dist = site_hazard_info['scenario_data'][[item for item in site_hazard_info['scenario_data'].keys() if 'dist' in item][0]]

        fig, ax, export_mat_dict = Utility.plot_3d_bar(mag, dist, rate,
                                                       figkwargs=figkwargs, cmap=cmap,
                                                       grid_alpha=grid_alpha, minor_grid=minor_grid,
                                                       fig_ax=fig_ax)

        export_mat_dict['mag'] = export_mat_dict.pop('x')
        export_mat_dict['dist'] = export_mat_dict.pop('y')
        export_mat_dict['rate'] = export_mat_dict.pop('dz')

        if save_mat:
            name = self.name
            file_name = f'plot_scenario_rates_{name}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    def plot_rv_distribution_functions(self, rv_index, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        k = kwargs.get('k', 5)
        n_cont = kwargs.get('n_cont', 100)
        plot_cdf = kwargs.get('plot_cdf', True)
        color_pdf = kwargs.get('color_pdf', 'blue')
        color_cdf = kwargs.get('color_cdf', 'red')
        lw = kwargs.get('lw', 1)
        scilimits = kwargs.get('scilimits', (-4, 4))
        fig_ax = kwargs.get('fig_ax', None)

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig, ax1 = plt.subplots(**figkwargs)
            ax2 = ax1.twinx()
        else:
            fig = fig_ax[0]
            axs = fig_ax[1]
            ax1 = axs[0]
            if len(axs) == 2:
                ax2 = axs[1]
            else:
                ax2 = ax1.twinx()

        prob_dist = self.get_flattened_rv_prob_dist_list()[rv_index]
        dist_name = prob_dist.dist.name
        mean = prob_dist.mean()
        std = prob_dist.std()
        x1 = mean - k * std
        x2 = mean + k * std
        x = np.linspace(x1, x2, n_cont)
        pdf_x = prob_dist.pdf(x)
        cdf_x = prob_dist.cdf(x)

        ax1.ticklabel_format(axis='both', scilimits=scilimits)
        if plot_cdf:
            ax1.minorticks_on()
            ax1.grid(True, which="major", alpha=grid_alpha, color=color_pdf)
            ax1.grid(True, which="minor", alpha=minor_grid_alpha, color=color_pdf)
            ax1.tick_params(axis='y', labelcolor=color_pdf)
            ax2.minorticks_on()
            ax2.grid(True, which="major", alpha=grid_alpha, color=color_cdf)
            ax2.grid(True, which="minor", alpha=minor_grid_alpha, color=color_cdf)
            ax2.tick_params(axis='y', labelcolor=color_cdf)
            ax2.plot(x, cdf_x, color=color_cdf, linewidth=lw)
            axs = [ax1, ax2]
        else:
            ax1.minorticks_on()
            ax1.grid(True, which="major", alpha=grid_alpha, color='black')
            ax1.grid(True, which="minor", alpha=minor_grid_alpha, color='black')
            ax1.tick_params(axis='y', labelcolor='black')
            axs = [ax1]
            ax2.axis('off')

        ax1.plot(x, pdf_x, color=color_pdf, linewidth=lw)

        export_mat_dict = dict()
        export_mat_dict['x'] = x
        export_mat_dict['pdf_x'] = pdf_x
        export_mat_dict['cdf_x'] = cdf_x
        export_mat_dict['dist_name'] = dist_name

        if save_mat:
            name = self.name
            file_name = f'plot_rv_distribution_functions_{name}_rv_{rv_index}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, axs, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    def plot_random_model_param_realizations(self, for_which, rv_indices, **kwargs):
        rng_seed_recvd = kwargs.get('rng_seed', None)
        analysis_case = for_which[Structure.get_dir_level_index('Analysis_Case')]
        sample_size = kwargs.get('sample_size', 1000)
        sampling_method = kwargs.get('sampling_method', 'mcs')
        dist_func = kwargs.get('dist_func', 'pdf')

        prob_dist_list, corr_matrix = self.get_prob_dist_list_from_flattened(rv_indices)
        m = multivariate_nataf(prob_dist_list, corr_matrix)
        data = None

        if 'Deterministic' not in Structure.get_analysis_type(analysis_case):

            if len(for_which) > Structure.get_dir_level_index('Design_Num') + 1:
                dir_level_to_seek = 'Model_Option_Num'
                for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
                realizations = self.get_random_model_param_vals(for_which)[:, :, int(for_which[-1]) - 1]
                data = realizations[:, rv_indices]

            if analysis_case in Structure.get_analysis_cases('Incl_FE_Model_Param_Uncertainty'):
                if data is None:
                    data = m.rvs(sample_size, rng_seed_recvd, sampling_method)

            elif analysis_case in Structure.get_analysis_cases('Incl_Prob_Dist_Param_Estimation_Uncertainty'):
                itr_group = 0
                if rng_seed_recvd is None:
                    rng_seed = None
                else:
                    rng_seed = Utility.get_num([rng_seed_recvd, itr_group])

                prob_dist_list_temp = list()
                num_rv = len(prob_dist_list)
                dist_param_vals_list, corr_matrix_realztns = Utility.generate_random_dist_param_vals(
                    prob_dist_list,
                    corr_matrix,
                    self.model_params['random_model_params']['estimation_sample_size'],
                    self.model_params['random_model_params']['dist_params_sample_size'],
                    rng_seed=rng_seed
                )
                for iter_num_rv in range(num_rv):
                    prob_dist_list_temp.append(
                        mixture(dist_param_vals_list[iter_num_rv], each=prob_dist_list[iter_num_rv].dist))
                m = multivariate_nataf(prob_dist_list_temp, corr_matrix_realztns)
                prob_dist_list = prob_dist_list_temp

                if data is None:
                    data = m.rvs(sample_size, rng_seed_recvd, sampling_method)
            else:
                return

            if data is None:
                return

            figkwargs = kwargs.get('figkwargs', dict())
            grid_alpha = kwargs.get('grid_alpha', 0.5)
            minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
            save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
            save_mat = kwargs.get('save_mat', False)
            k = kwargs.get('k', 5)
            n_cont = kwargs.get('n_cont', 100)
            n_bins = kwargs.get('n_bins', 10)
            ms = kwargs.get('ms', 10)
            lc = kwargs.get('lc', 'black')
            lw = kwargs.get('lw', 1)
            hc = kwargs.get('hc', 'gray')
            mc = kwargs.get('mc', 'red')
            same_lim = kwargs.get('same_lim', False)
            fig_ax = kwargs.get('fig_ax', None)

            if len(rv_indices) == 1:
                if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
                    fig = plt.figure(**figkwargs)
                    axs = [fig.gca()]
                else:
                    fig = fig_ax[0]
                    axs = fig_ax[1]
                fig, axs[0], export_mat_dict = Utility.scatter_hist_1d(data, prob_dist_list[0],
                                                                       figkwargs=figkwargs,
                                                                       grid_alpha=grid_alpha,
                                                                       minor_grid_alpha=minor_grid_alpha,
                                                                       k=k, n_cont=n_cont,
                                                                       n_bins=n_bins, ms=ms, mc=mc,
                                                                       lw=lw, lc=lc, hc=hc,
                                                                       dist_func=dist_func,
                                                                       fig_ax=[fig, axs[0]])

                export_mat_dict[f'{dist_func}_plot_1'] = export_mat_dict.pop(f'{dist_func}_plot')
                export_mat_dict['dist_name_1'] = export_mat_dict.pop(f'dist_name')

            elif len(rv_indices) == 2:
                fig, axs, export_mat_dict = Utility.scatter_hist_2d(data, m,
                                                                    figkwargs=figkwargs,
                                                                    grid_alpha=grid_alpha,
                                                                    minor_grid_alpha=minor_grid_alpha,
                                                                    k=k, n_cont=n_cont,
                                                                    n_bins=n_bins, ms=ms, mc=mc,
                                                                    lw=lw, lc=lc, hc=hc,
                                                                    same_lim=same_lim, dist_func=dist_func,
                                                                    fig_ax=fig_ax)
                if sampling_method == 'lhs':
                    axs[0].grid(False, which="both")
                    grid, ind = m.get_2d_lhs_grid(sample_size, n_cont)
                    grid = grid[-1]
                    axs[0] = Utility.plot_grid(axs[0], grid[:, :, 0], grid[:, :, 1], ind, lw, lc, grid_alpha)

            elif len(rv_indices) == 3:
                if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
                    fig = plt.figure(**figkwargs)
                    axs = list()
                    axs.append(fig.add_subplot(2, 2, 1, projection='3d'))
                    axs.append(fig.add_subplot(2, 2, 3))
                    axs.append(fig.add_subplot(2, 2, 4))
                    axs.append(fig.add_subplot(2, 2, 2))
                else:
                    fig = fig_ax[0]
                    axs = fig_ax[1]

                if minor_grid_alpha > 0:
                    axs[0].minorticks_on()
                else:
                    axs[0].minorticks_off()
                axs[0].plot(data[:, 0], data[:, 1], data[:, 2], '.', markersize=ms, color=mc)

                export_mat_dict = dict()
                axes_lim = list()
                for itr in range(len(prob_dist_list)):
                    fig, axs[itr + 1], export_mat_dict = Utility.scatter_hist_1d(data[:, itr], prob_dist_list[itr],
                                                                                 grid_alpha=grid_alpha,
                                                                                 minor_grid_alpha=minor_grid_alpha,
                                                                                 k=k, n_cont=n_cont,
                                                                                 n_bins=n_bins, ms=ms, mc=mc,
                                                                                 lw=lw, lc=lc, hc=hc,
                                                                                 dist_func=dist_func,
                                                                                 fig_ax=[fig, axs[itr + 1]])
                    axes_lim.append(axs[itr + 1].get_xlim())
                    export_mat_dict[f'{dist_func}_plot_{itr + 1}'] = export_mat_dict.pop(f'{dist_func}_plot')
                    export_mat_dict[f'dist_name_{itr + 1}'] = export_mat_dict.pop(f'dist_name')

                axes_lim = np.array(axes_lim)
                if same_lim:
                    axes_lim[0:, :] = np.array([np.min(axes_lim[:, 0]), np.max(axes_lim[:, 1])])
                axs[0].set_xlim(axes_lim[0, :])
                axs[0].set_ylim(axes_lim[1, :])
                axs[0].set_zlim(axes_lim[2, :])
                axs[1].set_xlim(axes_lim[0, :])
                axs[2].set_xlim(axes_lim[1, :])
                axs[3].set_xlim(axes_lim[2, :])

                export_mat_dict[f'data'] = data

            else:
                return

            if save_mat:
                name = self.name
                for_which_str = '_'.join(for_which)
                file_name = f"plot_random_model_param_realizations_{name}_{for_which_str}_" \
                            f"rv_{'_'.join(map(str, rv_indices))}.mat"
                sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

            return fig, axs, export_mat_dict

        else:
            return

    # ------------------------------------------------------------------------------------------------------------------

    def get_mixture_prob_dist_list(self, rv_indices, rng_seed=None):
        prob_dist_list, corr_matrix = self.get_prob_dist_list_from_flattened(rv_indices)
        dist_param_vals_list, corr_matrix_realztns = Utility.generate_random_dist_param_vals(
            prob_dist_list,
            corr_matrix,
            self.model_params['random_model_params']['estimation_sample_size'],
            self.model_params['random_model_params']['dist_params_sample_size'],
            rng_seed=rng_seed
        )
        num_rv = len(prob_dist_list)
        prob_dist_list_temp = list()
        for iter_num_rv in range(num_rv):
            prob_dist_list_temp.append(
                mixture(dist_param_vals_list[iter_num_rv], each=prob_dist_list[iter_num_rv].dist))
        return prob_dist_list_temp, multivariate_nataf(prob_dist_list_temp, corr_matrix_realztns)

    # ------------------------------------------------------------------------------------------------------------------

    def create_model_options(self, analysis_case):
        model_attributes = self.model_params['model_attributes']
        if 'Incl_Model_Form_Error' not in Structure.get_analysis_type(analysis_case):
            return [[item[0] for item in list(model_attributes.values())]]
        opts = 1
        opt_inds = list()
        model_attr_vals = list(model_attributes.values())
        for item in model_attr_vals:
            opt_inds.append(list(range(len(item))))
            opts *= len(item)

        cases = Utility.get_multi_level_iterator_cases(*opt_inds).astype(int)
        model_opts = list()
        for opt in range(opts):
            model_opts.append([model_attr_vals[itr][cases[opt, itr]] for itr in range(len(model_attr_vals))])
        return model_opts

    # ------------------------------------------------------------------------------------------------------------------

    def get_model_option_num_list(self, analysis_case, model_attributes):
        model_opts = self.create_model_options(analysis_case)
        inds = np.arange(len(model_opts), dtype=int)
        attr_keys = list(self.model_params['model_attributes'].keys())
        to_ret = list()
        for key in model_attributes.keys():
            for ind in inds:
                if model_opts[ind][attr_keys.index(key)] in model_attributes[key] and (ind + 1) not in to_ret:
                    to_ret.append(ind + 1)
        return to_ret

    # ------------------------------------------------------------------------------------------------------------------

    def fetch_random_model_param_vals(self, for_which, **kwargs):
        random_model_param_vals = kwargs.get('random_model_param_vals', self.get_random_model_param_vals(for_which))
        for_which = Structure.get_modified_for_which(for_which, 'Model_Realization_Num')
        iter_model_realztn = int(for_which[Structure.get_dir_level_index('Model_Realization_Num')])
        iter_model_option = int(for_which[Structure.get_dir_level_index('Model_Option_Num')])
        return random_model_param_vals[iter_model_realztn - 1, :, iter_model_option - 1]

    def write_model_param_vals(self, dir_category, for_which, **kwargs):
        random_model_param_vals = kwargs.get('random_model_param_vals', self.get_random_model_param_vals(for_which))
        model_param_vals = self.fetch_random_model_param_vals(for_which,
                                                              random_model_param_vals=random_model_param_vals)
        for_which = Structure.get_modified_for_which(for_which, 'Model_Realization_Num')
        design_num = for_which[Structure.get_dir_level_index('Design_Num')]
        name_value_pairs = [
            (self.get_flattened_rv_name_list(), model_param_vals),
            (self.model_params['primary_design_params']['name_list'], self.model_params['primary_design_params']['value_list_dict'][design_num]),
            (self.model_params['other_model_params']['name_list'], self.model_params['other_model_params']['value_list']),
        ]
        work_dir_path = Utility.get_path(self.model_work_dir_path, 'Work_Dir', dir_category, *for_which)
        self.structural_analysis_platform.write_model_param_vals(work_dir_path, name_value_pairs)

    # ------------------------------------------------------------------------------------------------------------------

    def fetch_model_attributes(self, for_which, **kwargs):
        model_attributes_values = kwargs.get('model_attributes_values', self.create_model_options(for_which[0]))
        for_which = Structure.get_modified_for_which(for_which, 'Model_Option_Num')
        iter_model_option = int(for_which[Structure.get_dir_level_index('Model_Option_Num')])
        return dict(zip(self.model_params['model_attributes'].keys(), model_attributes_values[iter_model_option - 1]))

    def write_model_attributes(self, dir_category, for_which, **kwargs):
        model_attributes_values = kwargs.get('model_attributes_values', self.create_model_options(for_which[0]))
        name_value_pairs = self.fetch_model_attributes(for_which, model_attributes_values=model_attributes_values)
        name_value_pairs = [
            (list(name_value_pairs.keys()), list(name_value_pairs.values()))
        ]
        for_which = Structure.get_modified_for_which(for_which, 'Model_Realization_Num')
        work_dir_path = Utility.get_path(self.model_work_dir_path, 'Work_Dir', dir_category, *for_which)
        self.structural_analysis_platform.write_model_attributes(work_dir_path, name_value_pairs)

    # ------------------------------------------------------------------------------------------------------------------

    def get_damping_model(self, for_which):
        damping_model_name = self.fetch_model_attributes(for_which)["damping"]
        damping_models = self.model_params['damping_models']
        damping_model = [item for item in damping_models if item['name'] == damping_model_name][0]
        return damping_model

    def get_rayleigh_damping_params(self, for_which, damping_model=None):
        if damping_model is None:
            damping_model = self.get_damping_model(for_which)
        if damping_model['name'] == 'rayleigh_damping':
            xi_i = damping_model['xi_i']
            i = damping_model['i']
            w_i = damping_model['w_i']
            rmpvs = self.fetch_random_model_param_vals(for_which)
            damp_ratio_vals = list()
            for xi in xi_i:
                if type(xi) is str:
                    rv_index = self.get_rv_index(xi, whole=True)[0]
                    damp_ratio_vals.append(rmpvs[rv_index])
                else:
                    damp_ratio_vals.append(xi)
            if i[0] is not None:
                w_i[0] = 2 * np.pi / self.get_periods(for_which, i[0])
            if i[1] is not None:
                w_i[1] = 2 * np.pi / self.get_periods(for_which, i[1])
            alpha, beta = Utility.get_rayleigh_damping_params(damp_ratio_vals[0], w_i[0], damp_ratio_vals[1], w_i[1])
            return alpha, beta
        else:
            raise ValueError(f"Damping model used in for_which = {for_which} is not of type Rayleigh!")

    def get_modal_damping_params(self, for_which, damping_model=None):
        if damping_model is None:
            damping_model = self.get_damping_model(for_which)
        if damping_model['name'] == 'modal_damping':
            xi_i = damping_model['xi_i']
            rmpvs = self.fetch_random_model_param_vals(for_which)
            damp_ratio_vals = list()
            for xi in xi_i:
                if type(xi) is str:
                    rv_index = self.get_rv_index(xi, whole=True)[0]
                    damp_ratio_vals.append(rmpvs[rv_index])
                else:
                    damp_ratio_vals.append(xi)
            return len(xi_i), damp_ratio_vals
        else:
            raise ValueError(f"Damping model used in for_which = {for_which} is not of type Modal!")

    def plot_rayleigh_damping_model(self, for_which, freq, freq_type, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        lc = kwargs.get('lc', 'black')
        lw = kwargs.get('lw', 1.0)
        mc = kwargs.get('mc', 'red')
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        fig_ax = kwargs.get('fig_ax', None)

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)  # also accepts tight_layout=bool
            ax = fig.gca()
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        ax.minorticks_on()
        ax.grid(True, which="major", alpha=grid_alpha)
        ax.grid(True, which="minor", alpha=minor_grid_alpha)

        periods = self.get_periods(for_which, [1, 2, 3, 4, 5, 6, 7, 8])
        freq_periods = 2 * np.pi / periods
        if freq_type == 'circular':
            freq_input = 2 * np.pi * np.array(freq)
            freq_periods_plot = 1 / periods
        elif freq_type == 'angular':
            freq_input = np.array(freq)
            freq_periods_plot = freq_periods
        else:
            raise ValueError(f"freq_type should be in ['circular', 'angular']")

        alpha, beta = self.get_rayleigh_damping_params(for_which)
        ax.plot(freq, alpha / 2 / freq_input + beta * freq_input / 2, color=lc, linewidth=lw)
        ax.plot(freq_periods_plot, alpha / 2 / freq_periods + beta * freq_periods / 2, '*', color=mc)

        export_mat_dict = dict()
        export_mat_dict['freq'] = freq
        export_mat_dict['freq_periods_plot'] = freq_periods_plot
        export_mat_dict['freq_type'] = freq_type
        export_mat_dict['alpha'] = alpha
        export_mat_dict['beta'] = beta

        if save_mat:
            name = self.name
            for_which_str = '_'.join(for_which)
            file_name = f'plot_rayleigh_damping_model_{name}_{for_which_str}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    def write_damping_file(self, for_which):
        model_work_dir_path = self.model_work_dir_path
        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        dir_category = 'Prelim_Analysis_Results'
        work_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which))
        try:
            damping_model = self.get_damping_model(for_which)
            if 'rayleigh_damping' in damping_model['name']:
                alpha, beta = self.get_rayleigh_damping_params(for_which, damping_model)
                self.structural_analysis_platform.write_damping_file('rayleigh_damping', work_dir_path, alpha=alpha,
                                                                     beta=beta)
            elif 'modal_damping' in damping_model['name']:
                n_modes, damp_ratios = self.get_modal_damping_params(for_which, damping_model)
                self.structural_analysis_platform.write_damping_file('modal_damping', work_dir_path, n_modes=n_modes,
                                                                     damp_ratios=damp_ratios)
            else:
                raise NotImplementedError(f"{damping_model['name']} is NOT an implemented damping model.")
        except KeyError:
            print(f"Damping information not found for_which = {for_which}! Proceeding with empty damping file.")
            self.structural_analysis_platform.write_damping_file(None, work_dir_path)
            pass

    # ------------------------------------------------------------------------------------------------------------------

    def get_m_alpha(self, d_line_list):
        d_vals = np.array(self.get_design_pts(d_line_list))
        line_coeffs = np.polyfit(d_vals[:, 0], d_vals[:, 1], 1)
        m, alpha = line_coeffs[:]
        return m, alpha

    # ------------------------------------------------------------------------------------------------------------------

    def get_d_from_x(self, x, d_line_list):
        scalar = True if np.isscalar(x) else False
        if scalar:
            x = np.array([x])
        else:
            x = np.array(x)
        m, alpha = self.get_m_alpha(d_line_list)
        d = list()
        for itr in range(len(x)):
            d.append((np.linalg.inv(np.array([[1 / m, 1], [-m, 1]])) @ np.array([[x[itr]], [alpha]])).flatten())
        return np.array(d).flatten() if scalar else np.array(d)

    # ------------------------------------------------------------------------------------------------------------------

    def get_x_from_d(self, d, d_line_list):
        d = np.array(d)
        scalar = True if d.ndim == 1 else False
        if scalar:
            d = d[np.newaxis, :]
        m, alpha = self.get_m_alpha(d_line_list)
        x = d[:, 1] + 1 / m * d[:, 0]
        return x.item() if scalar else x

    # ------------------------------------------------------------------------------------------------------------------

    def triangulate_regular_grid_design_space(self):
        model_params = self.model_params
        points = model_params['primary_design_params']['value_list_dict']
        dp_qual = model_params['primary_design_params']['design_point_qualifier_dict']
        points = np.array([points[key] for key in points.keys() if 'non-gp' not in dp_qual[key]])

        x_points = np.sort(np.unique(points[:, 0]))
        y_points = np.sort(np.unique(points[:, 1]))
        x_len = len(x_points) - 1
        y_len = len(y_points) - 1
        grid_points = np.array([[x, y] for x in x_points for y in y_points])
        a = [[i + j * (y_len + 1), (i + 1) + j * (y_len + 1), i + (j + 1) * (y_len + 1)]
             for i in range(y_len) for j in range(x_len)]
        b = [[i + (j + 1) * (y_len + 1), (i + 1) + j * (y_len + 1), (i + 1) + (j + 1) * (y_len + 1)]
             for i in range(y_len) for j in range(x_len)]
        # noinspection PyTypeChecker
        triangulation = tri.Triangulation(grid_points[:, 0],
                                          grid_points[:, 1],
                                          triangles=np.array(a + b))
        return triangulation, grid_points

    # ------------------------------------------------------------------------------------------------------------------

    def plot_design_space(self, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        lw = kwargs.get('lw', 1)
        lc = kwargs.get('lc', 'black')
        mc_gp = kwargs.get('mc_gp', 'red')
        ms_gp = kwargs.get('ms_gp', 6)
        mt_gp = kwargs.get('mt_gp', 'o')
        mc_ngp = kwargs.get('mc_ngp', 'red')
        ms_ngp = kwargs.get('ms_ngp', 6)
        mt_ngp = kwargs.get('mt_ngp', 'd')
        mc_ad = kwargs.get('mc_ad', 'red')
        ms_ad = kwargs.get('ms_ad', 18)
        mt_ad = kwargs.get('mt_ad', '*')
        mc_mps = kwargs.get('mc_mps', list())
        ms_mps = kwargs.get('ms_mps', list())
        mt_mps = kwargs.get('mt_mps', list())
        mark_points = kwargs.get('mark_points', list())
        fig_ax = kwargs.get('fig_ax', None)

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)
            ax = fig.gca()
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        ax.minorticks_on()
        ax.grid(True, which="major", alpha=grid_alpha)
        ax.grid(True, which="minor", alpha=minor_grid_alpha)

        tr, grid_data = self.triangulate_regular_grid_design_space()
        ax.set_xticks(np.unique(grid_data[:, 0]))
        ax.set_yticks(np.unique(grid_data[:, 1]))
        dp_qual = self.model_params['primary_design_params']['design_point_qualifier_dict']
        as_designed_key = [key for key in dp_qual.keys() if 'as-designed' in dp_qual[key]]
        non_gp_keys = [key for key in dp_qual.keys() if 'non-gp' in dp_qual[key] and 'as-designed' not in dp_qual[key]]
        as_designed = self.model_params['primary_design_params']['value_list_dict'][as_designed_key[0]]
        non_gp_dps = np.array(
            [self.model_params['primary_design_params']['value_list_dict'][key] for key in non_gp_keys])
        triangles = tr.triangles

        # for itr in range(triangles.shape[0]):
        #     ax.plot(grid_data[np.append(triangles[itr], triangles[itr][0]), 0],
        #             grid_data[np.append(triangles[itr], triangles[itr][0]), 1], '-', lw=lw, color=lc)

        ax.triplot(tr, lw=lw, color=lc)

        ax.plot(grid_data[:, 0], grid_data[:, 1], mt_gp,
                ms=ms_gp, markerfacecolor=mc_gp, markeredgecolor='black', markeredgewidth=0.5)
        ax.plot(*as_designed, mt_ad, ms=ms_ad, markerfacecolor=mc_ad, markeredgecolor='black', markeredgewidth=0.5)
        if non_gp_dps.size != 0:
            ax.plot(non_gp_dps[:, 0], non_gp_dps[:, 1], mt_ngp, ms=ms_ngp, markerfacecolor=mc_ngp,
                    markeredgecolor='black', markeredgewidth=0.5)

        for itr in range(len(mark_points)):
            ax.plot(*mark_points[itr], mt_mps[itr], ms=ms_mps[itr], markerfacecolor=mc_mps[itr],
                    markeredgecolor='black', markeredgewidth=0.5)

        dx = np.mean(np.diff(np.unique(grid_data[:, 0])))
        dy = np.mean(np.diff(np.unique(grid_data[:, 1])))

        ax.set_xlim([np.min(grid_data[:, 0]) - dx, np.max(grid_data[:, 0]) + dx])
        ax.set_ylim([np.min(grid_data[:, 1]) - dy, np.max(grid_data[:, 1]) + dy])

        export_mat_dict = dict()
        export_mat_dict['triangles'] = triangles
        export_mat_dict['grid_data'] = grid_data
        export_mat_dict['as_designed'] = as_designed

        if save_mat:
            file_name = f'plot_design_space_{self.name}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    def extract_prelim_analysis(self, for_which, extract_dest_path, num_modes):
        model_work_dir_path = self.model_work_dir_path
        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        dir_category = 'Prelim_Analysis_Results'
        work_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which))
        self.write_damping_file(for_which)
        prelim_analysis_files = self.structural_analysis_platform.get_prelim_analysis_files()
        for file in prelim_analysis_files:
            shutil.copy(Utility.get_path(work_dir_path, file), extract_dest_path)
        temp_platform = deepcopy(self.structural_analysis_platform)
        temp_platform.model_files_path = extract_dest_path
        temp_platform.write_run_prelim_analysis(extract_dest_path, num_modes)

    def extract_nltha(self, for_which, extract_dest_path):
        model_work_dir_path = self.model_work_dir_path
        dir_category = 'NLTHA_Results'
        dir_level_to_seek = 'Ground_Motion_Rec_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        work_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which))
        nltha_files = self.structural_analysis_platform.get_nltha_files()
        for file in nltha_files:
            shutil.copy(Utility.get_path(work_dir_path, file), extract_dest_path)
        temp_platform = deepcopy(self.structural_analysis_platform)
        temp_platform.model_files_path = extract_dest_path
        temp_platform.write_run_nltha(
            extract_dest_path,
            extract_dest_path,
            write_model_files=1,
            generate_additional_recorders=True,
            delete_rec=False,
        )


class FrameStructure(Structure):
    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, name, location_info, model_files_path, model_work_dir_path, model_params, structural_analysis_platform):
        super().__init__(name, location_info, model_files_path, model_work_dir_path, model_params, structural_analysis_platform)

    @staticmethod
    def read_secdef_file(file_path, delim):
        with open(file_path, 'r') as section_file_fid:
            section_file_lines = [[item.strip(', "\'') for item in line.strip().strip(';').split(delim)]
                                  for line in section_file_fid.readlines() if not line.isspace()]
        return section_file_lines

    @staticmethod
    def get_mat_tags_from_secdef(secdef):
        mat_tags = []
        for item in secdef:
            if item[0] in ['patch', 'layer']:
                mat_tags.append(int(item[2]))
            if item[0] == 'fiber':
                mat_tags.append(int(item[-1]))
        return mat_tags

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def generate_rebar_stress_strain_recorder(self, col_id, edge_id, col_edge_sec_info, rec_save_dir_path, rec_gen_file_fid):
        section_file_lines = col_edge_sec_info['secdef']
        ele_tag = col_edge_sec_info['ele_tag']
        sec_num = col_edge_sec_info['sec_num']
        fib_ctr = 0
        for curr_line in section_file_lines:
            first_word = curr_line[0]

            if first_word == 'fiber':
                # code block
                pass
            if first_word == 'patch':
                # code block
                pass
            if first_word == 'layer':
                second_word = curr_line[1]

                if second_word == 'straight':
                    # code block
                    pass
                if second_word == 'circ':
                    mat_tag = int(curr_line[2])
                    if mat_tag in col_edge_sec_info['rebar_mat_tags']:
                        fib_ctr = self.write_circ_layer_stress_strain_recorder(
                            curr_line, ele_tag, sec_num, col_id, edge_id, fib_ctr, rec_save_dir_path, rec_gen_file_fid)
        return

    # ------------------------------------------------------------------------------------------------------------------

    def write_circ_layer_stress_strain_recorder(self, curr_line, ele_tag, sec_num, col_id, edge_id, fib_ctr,
                                                rec_save_dir_path, rec_gen_file_fid):
        mat_tag = curr_line[2]
        num_fiber = int(curr_line[3])
        y_center = float(curr_line[5])
        z_center = float(curr_line[6])
        radius = float(curr_line[7])
        if len(curr_line) > 8:
            start_ang = float(curr_line[8])
            end_ang = float(curr_line[9])
        else:
            start_ang = 0.0
            end_ang = 360.0 - 360.0 / num_fiber

        ang = start_ang
        ang_incr = (end_ang - start_ang) / (num_fiber - 1)

        # Write recorders for rebars of current layer
        for i_fib in range(num_fiber):
            fib_ctr += 1
            y = y_center + radius * np.cos(np.deg2rad(ang))
            z = z_center + radius * np.sin(np.deg2rad(ang))
            ang += ang_incr
            recorder_file_name = f'fib_stress_strain_col_{col_id}_edge_{edge_id}_elem_{ele_tag}' \
                                 f'_mat_{mat_tag}_fib_{fib_ctr}.txt'
            recorder_file_path = Utility.get_path(rec_save_dir_path, recorder_file_name)
            new_recorder_line = self.structural_analysis_platform.get_stress_strain_recorder_command(
                recorder_file_path, ele_tag, sec_num, y, z, mat_tag)
            rec_gen_file_fid.write(new_recorder_line)
        return fib_ctr

    def write_col_rebar_strain_recorder(self, for_which, work_dir_path,
                                        rec_gen_file_open_mode, rec_save_dir_path):
        rec_gen_file_fid = self.structural_analysis_platform.open_edp_recorder_file(
            work_dir_path,
            rec_gen_file_open_mode,
            rec_save_dir_path
        )
        col_edge_sec_info = self.get_col_edge_sec_info(for_which)
        num_cols_total = self.model_params['num_cols_total']
        col_edge_list = self.model_params['col_edge_list']

        for i_col in range(1, num_cols_total + 1):
            col_id = i_col
            for i_edge in col_edge_list:
                edge_id = i_edge

                self.generate_rebar_stress_strain_recorder(
                    col_id, edge_id,
                    col_edge_sec_info[col_id, edge_id],
                    rec_save_dir_path,
                    rec_gen_file_fid,
                )
        rec_gen_file_fid.close()

    def get_col_edge_sec_info(self, for_which):
        dir_category = 'Prelim_Analysis_Results'
        for_which = Structure.get_modified_for_which(for_which, 'Model_Realization_Num')
        prelim_analysis_work_dir_path = Utility.get_path(self.model_work_dir_path, 'Work_Dir', dir_category, *for_which)
        model_info_files_dir_path = Utility.get_path(prelim_analysis_work_dir_path, self.structural_analysis_platform.model_info_dir_name)

        col_rebar_mat_tags = self.get_col_rebar_mat_tags(for_which)

        col_edge_info = {}
        num_cols_total = self.model_params['num_cols_total']
        col_edge_list = self.model_params['col_edge_list']

        for col_id in range(1, num_cols_total + 1):
            col_elem_sec_info = self.get_col_elem_sec_info(for_which, col_id)
            for edge_id in col_edge_list:
                dict_key = (col_id, edge_id)

                ele_tag, sec_tag, sec_num = None, None, None
                if len(col_elem_sec_info) == 1 and edge_id == 1:
                    ele_tag = col_elem_sec_info[0][0]
                    sec_tag = col_elem_sec_info[0][1]
                    sec_num = 1

                if len(col_elem_sec_info) == 1 and edge_id == 2:
                    ele_tag = col_elem_sec_info[0][0]
                    sec_tag = col_elem_sec_info[0][-1]
                    sec_num = len(col_elem_sec_info[0]) - 1

                if len(col_elem_sec_info) > 1 and edge_id == 1:
                    ele_tag = col_elem_sec_info[0][0]
                    sec_tag = col_elem_sec_info[0][1]
                    sec_num = 1

                if len(col_elem_sec_info) > 1 and edge_id == 2:
                    ele_tag = col_elem_sec_info[-1][0]
                    sec_tag = col_elem_sec_info[-1][-1]
                    sec_num = len(col_elem_sec_info[-1]) - 1

                secdef_file_path = Utility.get_path(model_info_files_dir_path, f'fib_secdef_sec_{sec_tag}.txt')
                section_file_lines_space = FrameStructure.read_secdef_file(secdef_file_path, delim=' ')
                section_file_lines_comma = FrameStructure.read_secdef_file(secdef_file_path, delim=',')
                section_file_lines = None
                for itr in range(len(section_file_lines_comma)):
                    if len(section_file_lines_comma[itr]) > len(section_file_lines_space[itr]):
                        section_file_lines = section_file_lines_comma
                        break
                    elif len(section_file_lines_comma[itr]) < len(section_file_lines_space[itr]):
                        section_file_lines = section_file_lines_space
                        break
                    else:
                        pass
                if section_file_lines is None:
                    # both are same, choose any one
                    section_file_lines = section_file_lines_space

                section_mat_tags = FrameStructure.get_mat_tags_from_secdef(section_file_lines)

                col_edge_info[dict_key] = {
                    'ele_tag': ele_tag,
                    'sec_num': sec_num,
                    'secdef': section_file_lines,
                    'rebar_mat_tags': [mat_tag for mat_tag in col_rebar_mat_tags if mat_tag in section_mat_tags]
                }
        return col_edge_info

    def generate_spring_elem_deformation_recorder(self, ele_tag, spring_type, spring_id,
                                                  rec_save_dir_path, rec_gen_file_fid):
        recorder_file_name = f'spring_deformation_{spring_type}_{spring_id}_elem_{ele_tag}.txt'
        recorder_file_path = Utility.get_path(rec_save_dir_path, recorder_file_name)
        new_recorder_line = self.structural_analysis_platform.get_elem_deformation_recorder_command(recorder_file_path, ele_tag)
        rec_gen_file_fid.write(new_recorder_line)
        return

    def write_spring_deformation_recorder(self, for_which, spring_type, work_dir_path,
                                          rec_gen_file_open_mode, rec_save_dir_path):
        rec_gen_file_fid = self.structural_analysis_platform.open_edp_recorder_file(
            work_dir_path,
            rec_gen_file_open_mode,
            rec_save_dir_path
        )
        spring_elem_tags = self.get_spring_elem_tags(for_which, spring_type)
        for i_spring in range(1, len(spring_elem_tags) + 1):
            self.generate_spring_elem_deformation_recorder(
                spring_elem_tags[i_spring - 1], spring_type, i_spring, rec_save_dir_path, rec_gen_file_fid)
        rec_gen_file_fid.close()

    def get_col_rebar_mat_tags(self, for_which):

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        model_work_dir_path = self.model_work_dir_path
        dir_category = 'Prelim_Analysis_Results'
        file_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which),
                                     self.structural_analysis_platform.model_info_dir_name, 'col_rebar_mat_info.txt')

        col_rebar_mat_tags = Utility.read_numpy_array_from_txt_file(file_path, skiprows='find')
        return col_rebar_mat_tags.flatten().astype(int)  # 1-d array

    # ------------------------------------------------------------------------------------------------------------------

    def get_spring_elem_tags(self, for_which, spring_type):

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        model_work_dir_path = self.model_work_dir_path
        dir_category = 'Prelim_Analysis_Results'
        file_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which),
                                     self.structural_analysis_platform.model_info_dir_name, f'{spring_type}_spring_elem_tags.txt')

        spring_elem_tags = Utility.read_numpy_array_from_txt_file(file_path, skiprows='find')
        return spring_elem_tags.flatten().astype(int)  # 1-d array

    # ------------------------------------------------------------------------------------------------------------------

    def get_col_elem_sec_info(self, for_which, col_id):

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        model_work_dir_path = self.model_work_dir_path
        dir_category = 'Prelim_Analysis_Results'
        file_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which),
                                     self.structural_analysis_platform.model_info_dir_name, f'col_elem_sec_info_col_{col_id}.txt')
        with open(file_path, 'r') as fid:
            lines = [line for line in fid.readlines() if not line.isspace()]
        return [list(np.array(line.strip(';\n ').split()).astype(int)) for line in lines]


class OSB(FrameStructure):

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, name, location_info, model_files_path, model_work_dir_path, model_params, structural_analysis_platform):
        super().__init__(name, location_info, model_files_path, model_work_dir_path, model_params, structural_analysis_platform)

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def get_first_trans_mode_info(self, for_which, **kwargs):
        # NOTE:
        # BridgeCoordsInOpenSees = R*BridgeCoordsInLTV
        # BridgeCoordsInLTV = R'*BridgeCoordsInOpenSees
        # Assume L,T,V as basis
        # 1st row of R = OSB's global X in L,T,V
        # 2nd row of R = OSB's global Y in L,T,V
        # 3rd row of R = OSB's global Z in L,T,V

        crd_transf_matrix = kwargs.get('crd_transf_matrix', np.eye(3))

        model_work_dir_path = self.model_work_dir_path
        dir_category = 'Prelim_Analysis_Results'

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        work_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which))

        primary_nodes_deck = Utility.read_numpy_array_from_txt_file(
            Utility.get_path(work_dir_path, self.structural_analysis_platform.model_info_dir_name, 'primary_nodes_deck.txt'), skiprows='find').flatten()
        periods = Utility.read_numpy_array_from_txt_file(
            Utility.get_path(work_dir_path, self.structural_analysis_platform.model_info_dir_name, 'periods.txt'),
            skiprows='find').flatten()

        mode_shapes = list()
        for i_mode in range(len(periods)):
            temp = Utility.read_numpy_array_from_txt_file(
                Utility.get_path(work_dir_path, self.structural_analysis_platform.model_info_dir_name, f'mode_shape_{i_mode + 1}.txt'),
                skiprows='find')
            temp[:, 1:4] = (crd_transf_matrix.T @ temp[:, 1:4].T).T
            mode_shapes.append(temp)

        norm_mode = np.zeros((len(periods), 2))

        # Find norm of mode shapes in L and T directions
        for i_mode in range(len(periods)):
            norm_mode[i_mode, 0] = np.linalg.norm(mode_shapes[i_mode][:, 1])
            norm_mode[i_mode, 1] = np.linalg.norm(mode_shapes[i_mode][:, 2])

        # Isolate modes whose T-norm is greater than L-norm
        temp = np.divide(norm_mode[:, 1], norm_mode[:, 0]) > 1.0
        trans_mode_index = np.flatnonzero(temp)  # starting zero

        # Find location of primary nodes of deck in EigenVector matrix
        primary_nodes_deck_loc = np.flatnonzero(np.in1d(mode_shapes[0][:, 0], primary_nodes_deck) == 1)

        # Find T-disp of primary nodes of deck
        disp_trans_primary_nodes_deck = np.zeros((len(primary_nodes_deck_loc), len(trans_mode_index)))
        for i_mode in range(len(trans_mode_index)):
            for i_primary_nodes_deck in range(len(primary_nodes_deck)):
                disp_trans_primary_nodes_deck[i_primary_nodes_deck, i_mode] = mode_shapes[trans_mode_index[i_mode]][
                    primary_nodes_deck_loc[i_primary_nodes_deck]][2]

        # Sum the sign of T-disp of primary nodes of deck for each isolated T-mode
        temp = np.abs(np.sum(np.sign(disp_trans_primary_nodes_deck), axis=0))

        # First T-mode is the mode for which most primary nodes of deck are transversely displaced in the same direction
        # return transModeNum (= trans_mode_index + 1), trans_mode_period
        trans_mode_index = trans_mode_index[np.argmin(np.abs(temp - len(primary_nodes_deck)))]
        trans_mode_period = periods[trans_mode_index]
        return trans_mode_index + 1, trans_mode_period

    # ------------------------------------------------------------------------------------------------------------------

    def get_period_range_extremes(self, for_which, define_range):
        try:
            period_extremes = super().get_period_range_extremes(for_which, define_range)
            return period_extremes
        except NotImplementedError:
            dir_level_to_seek = 'Model_Realization_Num'
            for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
            period_extremes = list()
            for itr in define_range:
                if itr == 'T_1_trans':
                    _, period = self.get_first_trans_mode_info(for_which)
                else:
                    raise NotImplementedError(f"period getter not defined for {itr}")
                period_extremes.append(period)
            if len(period_extremes) == 1:
                period_extremes = [period_extremes[0], period_extremes[0]]
            return period_extremes

    # ------------------------------------------------------------------------------------------------------------------

    def get_rayleigh_damping_params(self, for_which, damping_model=None):
        try:
            alpha, beta = super().get_rayleigh_damping_params(for_which, damping_model)
            return alpha, beta
        except ValueError:
            if damping_model is None:
                damping_model = self.get_damping_model(for_which)
            if damping_model['name'] == 'rayleigh_damping_first_trans_mode':
                xi_i = damping_model['xi_i'][0]
                rmpvs = self.fetch_random_model_param_vals(for_which)
                if type(xi_i) is str:
                    rv_index = self.get_rv_index(xi_i, whole=True)[0]
                    damp_ratio_val = rmpvs[rv_index]
                else:
                    damp_ratio_val = xi_i
                alpha, beta = self.get_rayleigh_damping_first_trans_mode_params(for_which, damp_ratio_val)
                return alpha, beta
            else:
                raise ValueError(f"Damping model used in for_which = {for_which} is not of type Rayleigh!")

    # ------------------------------------------------------------------------------------------------------------------

    def get_rayleigh_damping_first_trans_mode_params(self, for_which, xi_1):
        _, damp_period = self.get_first_trans_mode_info(for_which)
        xi_2 = 1.10 * xi_1 if xi_1 > 0.05 else 0.05
        w_1 = 2 * np.pi / damp_period
        w_2 = 1.10 * xi_2 * w_1 / xi_1
        alpha, beta = Utility.get_rayleigh_damping_params(xi_1, w_1, xi_2, w_2)
        return alpha, beta

    # ------------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 12 14:01:56 2019

@author: Angshuman Deb
"""

from .analysis import Analysis
from .utility import Utility
from .structure import Structure
import os
import numpy as np
from time import time
from functools import partial
from pathos.multiprocessing import ProcessPool as Pool
from benedict import benedict


class PSHA(Analysis):

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, im, **kwargs):
        comp_env = kwargs.get('comp_env', 'local')
        self.im = im
        super().__init__(self.__class__.__name__, im.structure, comp_env)

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def run_parallel(setup_dir_path, compute_seismic_hazard_integral, im_input, node_index, batch_index,
                     analysis_index, run_case):
        start_time = time()

        for_which = list(run_case.astype(str))
        analysis_case = for_which[Structure.get_dir_level_index('Analysis_Case')]
        run_case_str = ' '.join(for_which)

        psha_results = compute_seismic_hazard_integral(for_which, im_input=im_input)

        status_file_name = f'status_analysis_index_{int(analysis_index)}' \
                           f'_node_{int(node_index)}_batch_{int(batch_index)}.txt'
        status_file_path = Utility.get_path(setup_dir_path, analysis_case, status_file_name)

        with open(status_file_path, 'w') as fid:
            fid.write(f"{run_case_str} : COMPLETED in {time() - start_time} seconds\n")

        return f'{run_case_str}', psha_results

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def stage(self, analysis_case, design_num_list):

        self.structure.set_site_hazard_info(self.im)

        results_dir_path = self.get_results_dir_path()

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)

            if not os.path.isdir(target_work_dir_path):
                os.makedirs(target_work_dir_path)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def setup(self, analysis_case, design_num_list, **kwargs):
        run_time_limit = kwargs.get('run_time_limit', '01:00:00')
        allocation_name = kwargs.get('allocation_name', '')
        n_batch = kwargs.get('n_batch', 1)
        n_job = kwargs.get('n_job', 1)
        pool_size_factor = kwargs.get('pool_size_factor', 1)

        structure = self.structure

        self.stage(analysis_case, design_num_list)

        setup_dir_path = self.get_setup_dir_path()
        if not os.path.isdir(Utility.get_path(setup_dir_path, analysis_case)):
            os.makedirs(Utility.get_path(setup_dir_path, analysis_case))

        temp = list()
        for design_num in design_num_list:
            temp.append(
                Utility.get_multi_level_iterator_cases(
                    int(analysis_case),
                    int(design_num),
                    list(range(1, structure.get_random_model_param_vals([analysis_case, design_num]).shape[2] + 1)),
                    list(range(1, structure.get_random_model_param_vals([analysis_case, design_num]).shape[0] + 1))
                )
            )
        analysis_list = np.vstack(temp)

        list_file_path = Utility.get_path(setup_dir_path, analysis_case, 'analysis_list.txt')
        Utility.save_array_as_text_file(list_file_path, analysis_list, fmt='%d',
                                        header=f'Total analyses to run = {analysis_list.shape[0]:d}')

        self.set_analysis_setup_info(analysis_case, {'design_num_list': design_num_list})

        if not self.comp_env == 'local':
            self.setup_non_local_run(analysis_case, n_batch, n_job, pool_size_factor, run_time_limit, allocation_name)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def run(self, analysis_case, pool_size, **kwargs):
        n_batch = kwargs.get('n_batch', 1)
        n_job = kwargs.get('n_job', 1)
        node_index = kwargs.get('node_index', 1)
        batch_index = kwargs.get('batch_index', 1)
        im_input = kwargs.get('im_input', np.array([]))

        compute_seismic_hazard_integral = self.im.compute_seismic_hazard_integral

        setup_dir_path = self.get_setup_dir_path()
        results_dir_path = self.get_results_dir_path()
        analysis_list, analysis_range, n_sim = self.get_analysis_info(analysis_case, 'analysis_list.txt',
                                                                      n_batch, n_job, node_index, batch_index)

        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))

        to_run = partial(PSHA.run_parallel, setup_dir_path, compute_seismic_hazard_integral, im_input,
                         node_index, batch_index)

        to_pass = [[analysis_index for analysis_index in analysis_range],
                   [analysis_list[analysis_index - 1, :] for analysis_index in analysis_range], ]

        if pool_size > 1:
            my_pool = Pool(pool_size)
            psha_results = dict(my_pool.map(to_run, *to_pass))
            my_pool.close()
            my_pool.join()
            my_pool.clear()
        else:
            psha_results = dict(list(map(to_run, *to_pass)))

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)
            psha_results_design_num = {key: psha_results[key] for key in psha_results.keys() if
                                       key.split()[Structure.get_dir_level_index('Design_Num')] == design_num}
            if n_batch == 1 and n_job == 1:
                Utility.pickle_dump_dict(Utility.get_path(target_work_dir_path, f'psha_results.pickle'),
                                         psha_results_design_num)
            else:
                Utility.pickle_dump_dict(
                    Utility.get_path(
                        target_work_dir_path, f'psha_results_node_{node_index}_batch_{batch_index}.pickle'
                    ), psha_results_design_num
                )

        # Write final_status_node_{}_batch_{}.txt
        contains_str = f'_node_{int(node_index)}_batch_{int(batch_index)}'
        Utility.combine_contents_of_files(Utility.get_path(setup_dir_path, analysis_case),
                                          [f'contains status_analysis_index_ {contains_str}'],
                                          f'final_status{contains_str}.txt',
                                          file_ext='.txt', delete_original=True)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def wrap_up(self, analysis_case):
        setup_dir_path = self.get_setup_dir_path()
        results_dir_path = self.get_results_dir_path()

        # Write final_status.txt
        Utility.combine_contents_of_files(Utility.get_path(setup_dir_path, analysis_case),
                                          ['startswith final_status_', 'startswith status_analysis_index_'],
                                          'final_status.txt',
                                          file_ext='.txt', delete_original=True)

        analysis_list, _, _ = self.get_analysis_info(analysis_case, 'analysis_list.txt', 1, 1, 1, 1)

        # Combine results from all nodes and batches
        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)
            files = os.listdir(target_work_dir_path)
            files = [file for file in files if file.startswith('psha_results') and file.endswith(
                '.pickle') and 'node' in file and 'batch' in file]
            if len(files) > 0:
                list_of_dicts = [Utility.pickle_load_dict(Utility.get_path(target_work_dir_path, file)) for file in
                                 files]

                psha_results_file_path = Utility.get_path(target_work_dir_path, 'psha_results.pickle')
                if os.path.isfile(psha_results_file_path):
                    psha_results = benedict(Utility.pickle_load_dict(psha_results_file_path))
                else:
                    psha_results = benedict()

                for d in list_of_dicts:
                    psha_results.merge(d)

                Utility.pickle_dump_dict(psha_results_file_path, psha_results)

                for file in files:
                    os.remove(Utility.get_path(target_work_dir_path, file))

        return

    # ------------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 19 12:11:31 2019

@author: Angshuman Deb
"""

from .analysis import Analysis
from .utility import Utility
from .structure import Structure
import os
import numpy as np
from time import time
from functools import partial
from pathos.multiprocessing import ProcessPool as Pool
from benedict import benedict


class GMS(Analysis):

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, im, **kwargs):
        comp_env = kwargs.get('comp_env', 'local')
        self.im = im
        super().__init__(self.__class__.__name__, im.structure, comp_env)

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def run_parallel(setup_dir_path, get_target_spectrum, select_ground_motion_records,
                     haz_lev_list, mrp_list, n_gm_list,
                     rng_seed, set_rng_seed, seed_multiplier,
                     uhs, deterministic, uncondition, max_scale, min_scale, dev_weights, n_loop, penalty, is_scaled,
                     not_allowed, classify_pulse, sampling_method, spectral_periods,
                     node_index, batch_index, analysis_index, run_case):

        for_which = list(run_case.astype(str))
        analysis_case = for_which[Structure.get_dir_level_index('Analysis_Case')]
        run_case_str = ' '.join(for_which)

        index = haz_lev_list.index(for_which[-1])
        mrp = mrp_list[index]
        n_gm = n_gm_list[index]

        if set_rng_seed:
            rng_seed_send = seed_multiplier * Utility.get_num(run_case[1:])
        else:
            rng_seed_send = rng_seed

        start_time = time()

        mean_req, cov_req, spectral_periods = get_target_spectrum(mrp, for_which,
                                                                  spectral_periods=spectral_periods,
                                                                  uncondition=uncondition,
                                                                  uhs=uhs)
        ground_motion_records = select_ground_motion_records(n_gm, mean_req, cov_req, spectral_periods, mrp, for_which,
                                                             deterministic=deterministic,
                                                             rng_seed=rng_seed_send,
                                                             max_scale=max_scale,
                                                             min_scale=min_scale,
                                                             dev_weights=dev_weights,
                                                             n_loop=n_loop,
                                                             penalty=penalty,
                                                             is_scaled=is_scaled,
                                                             not_allowed=not_allowed,
                                                             classify_pulse=classify_pulse,
                                                             sampling_method=sampling_method)

        gms_results = {'target_spectra': [mean_req, cov_req, spectral_periods],
                       'ground_motion_records': ground_motion_records,
                       'mrp': mrp,
                       'n_gm': n_gm}

        status_file_name = f'status_analysis_index_{int(analysis_index)}' \
                           f'_node_{int(node_index)}_batch_{int(batch_index)}.txt'
        status_file_path = Utility.get_path(setup_dir_path, analysis_case, status_file_name)

        with open(status_file_path, 'w') as fid:
            fid.write(f"{run_case_str} : COMPLETED in {time() - start_time} seconds\n")

        return f'{run_case_str}', gms_results

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def stage(self, analysis_case, design_num_list):

        results_dir_path = self.get_results_dir_path()

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)

            if not os.path.isdir(target_work_dir_path):
                os.makedirs(target_work_dir_path)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def setup(self, analysis_case, design_num_list, haz_lev_list, mrp_list, n_gm_list, **kwargs):
        run_time_limit = kwargs.get('run_time_limit', '01:00:00')
        allocation_name = kwargs.get('allocation_name', '')
        n_batch = kwargs.get('n_batch', 1)
        n_job = kwargs.get('n_job', 1)
        pool_size_factor = kwargs.get('pool_size_factor', 1)

        structure = self.structure

        self.stage(analysis_case, design_num_list)

        setup_dir_path = self.get_setup_dir_path()
        if not os.path.isdir(Utility.get_path(setup_dir_path, analysis_case)):
            os.makedirs(Utility.get_path(setup_dir_path, analysis_case))

        temp = list()
        for design_num in design_num_list:
            temp.append(
                Utility.get_multi_level_iterator_cases(
                    int(analysis_case),
                    int(design_num),
                    list(range(1, structure.get_random_model_param_vals([analysis_case, design_num]).shape[2] + 1)),
                    list(range(1, structure.get_random_model_param_vals([analysis_case, design_num]).shape[0] + 1)),
                    haz_lev_list
                )
            )
        analysis_list = np.vstack(temp)

        list_file_path = Utility.get_path(setup_dir_path, analysis_case, 'analysis_list.txt')
        Utility.save_array_as_text_file(list_file_path, analysis_list, fmt='%d',
                                        header=f'Total analyses to run = {analysis_list.shape[0]:d}')

        self.set_analysis_setup_info(
            analysis_case,
            {'design_num_list': design_num_list,
             'haz_lev_list': haz_lev_list,
             'mrp_list': mrp_list,
             'n_gm_list': n_gm_list
             }
        )

        if not self.comp_env == 'local':
            self.setup_non_local_run(analysis_case, n_batch, n_job, pool_size_factor, run_time_limit, allocation_name)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def run(self, analysis_case, pool_size, **kwargs):
        n_batch = kwargs.get('n_batch', 1)
        n_job = kwargs.get('n_job', 1)
        node_index = kwargs.get('node_index', 1)
        batch_index = kwargs.get('batch_index', 1)
        rng_seed = kwargs.get('rng_seed', None)
        uncondition = kwargs.get('uncondition', True)
        max_scale = kwargs.get('max_scale', 4)
        min_scale = kwargs.get('min_scale', 1 / 3)
        dev_weights = kwargs.get('dev_weights', [1, 2])
        n_loop = kwargs.get('n_loop', 0)
        penalty = kwargs.get('penalty', 0)
        is_scaled = kwargs.get('is_scaled', True)
        not_allowed = kwargs.get('not_allowed', [])
        classify_pulse = kwargs.get('classify_pulse', True)
        sampling_method = kwargs.get('sampling_method', 'mcs')
        spectral_periods = kwargs.get('spectral_periods', np.array([]))
        uhs = kwargs.get('uhs', False)
        deterministic = kwargs.get('deterministic', False)

        im = self.im
        get_target_spectrum = im.get_target_spectrum
        select_ground_motion_records = im.select_ground_motion_records

        setup_dir_path = self.get_setup_dir_path()
        results_dir_path = self.get_results_dir_path()
        analysis_list, analysis_range, n_sim = self.get_analysis_info(analysis_case, 'analysis_list.txt',
                                                                      n_batch, n_job, node_index, batch_index)

        analysis_setup_info = self.get_analysis_setup_info(analysis_case)
        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))
        haz_lev_list = analysis_setup_info['haz_lev_list']
        mrp_list = analysis_setup_info['mrp_list']
        n_gm_list = analysis_setup_info['n_gm_list']

        set_rng_seed, seed_multiplier = Utility.setup_rng(rng_seed)

        to_run = partial(GMS.run_parallel, setup_dir_path, get_target_spectrum, select_ground_motion_records,
                         haz_lev_list, mrp_list, n_gm_list,
                         rng_seed, set_rng_seed, seed_multiplier,
                         uhs, deterministic, uncondition, max_scale, min_scale, dev_weights, n_loop, penalty, is_scaled,
                         not_allowed, classify_pulse, sampling_method, spectral_periods,
                         node_index, batch_index)

        to_pass = [[analysis_index for analysis_index in analysis_range],
                   [analysis_list[analysis_index - 1, :] for analysis_index in analysis_range], ]

        if pool_size > 1:
            my_pool = Pool(pool_size)
            gms_results = dict(my_pool.map(to_run, *to_pass))
            my_pool.close()
            my_pool.join()
            my_pool.clear()
        else:
            gms_results = dict(list(map(to_run, *to_pass)))

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)
            gms_results_design_num = {key: gms_results[key] for key in gms_results.keys() if
                                      key.split()[Structure.get_dir_level_index('Design_Num')] == design_num}
            if n_batch == 1 and n_job == 1:
                Utility.pickle_dump_dict(
                    Utility.get_path(target_work_dir_path, f'gms_results.pickle'), gms_results_design_num)
            else:
                Utility.pickle_dump_dict(
                    Utility.get_path(
                        target_work_dir_path, f'gms_results_node_{node_index}_batch_{batch_index}.pickle'
                    ), gms_results_design_num
                )

        # Write final_status_node_{}_batch_{}.txt
        contains_str = f'_node_{int(node_index)}_batch_{int(batch_index)}'
        Utility.combine_contents_of_files(Utility.get_path(setup_dir_path, analysis_case),
                                          [f'contains status_analysis_index_ {contains_str}'],
                                          f'final_status{contains_str}.txt',
                                          file_ext='.txt', delete_original=True)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def wrap_up(self, analysis_case):
        setup_dir_path = self.get_setup_dir_path()
        results_dir_path = self.get_results_dir_path()

        # Write final_status.txt
        Utility.combine_contents_of_files(Utility.get_path(setup_dir_path, analysis_case),
                                          ['startswith final_status_', 'startswith status_analysis_index_'],
                                          'final_status.txt',
                                          file_ext='.txt', delete_original=True)

        analysis_list, _, _ = self.get_analysis_info(analysis_case, 'analysis_list.txt', 1, 1, 1, 1)

        # Combine results from all nodes and batches
        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)
            files = os.listdir(target_work_dir_path)
            files = [file for file in files if
                     file.startswith('gms_results') and file.endswith('.pickle') and 'node' in file and 'batch' in file]
            if len(files) > 0:
                list_of_dicts = [Utility.pickle_load_dict(Utility.get_path(target_work_dir_path, file))
                                 for file in files]

                gms_results_file_path = Utility.get_path(target_work_dir_path, 'gms_results.pickle')
                if os.path.isfile(gms_results_file_path):
                    gms_results = benedict(Utility.pickle_load_dict(gms_results_file_path))
                else:
                    gms_results = benedict()

                for d in list_of_dicts:
                    gms_results.merge(d)

                Utility.pickle_dump_dict(gms_results_file_path, gms_results)

                for file in files:
                    os.remove(Utility.get_path(target_work_dir_path, file))

        return

    # ------------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
"""
Created on Sun Nov 17 16:16:25 2019

@author: Angshuman Deb
"""

from .im import IM
import numpy as np
from .structure import Structure
from .utility import Utility
from .multivariate_nataf import multivariate_nataf
from .mixture import mixture
from abc import abstractmethod
import pygmm
from scipy.stats import lognorm
from scipy.stats import skew
import os
import scipy.io as sio
from scipy.stats import norm
import matplotlib.pyplot as plt
from matplotlib import colors
from numba import jit


class Sa(IM):
    large_err = 1.0e99

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, structure, gmm, correl_func):
        self.correl_func = correl_func  # function defining correlation b/w Sa at different structural periods
        super().__init__(structure, gmm)

    ####################################################################################################################
    # Abstract methods
    ####################################################################################################################

    @abstractmethod
    def setup_psa_spectra_gms(self, *args, **kwargs):
        pass

    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def _get_target_spectrum(self, *args, **kwargs):
        pass

    @abstractmethod
    def evaluate_period(self, for_which):
        pass

    @abstractmethod
    def mark_period_on_axis(self, for_which, ax, **kwargs):
        pass

    ####################################################################################################################
    # Public Functionalities (instance methods)
    ####################################################################################################################

    def shc_at_periods(self, periods, **kwargs):
        im_input = kwargs.get('im_input', np.array([]))
        im_input = np.array(im_input).flatten()
        if im_input.size == 0:
            im_input = np.logspace(np.log10(1e-4), np.log10(5), 1000, endpoint=True)
        med_sa_periods, sig_ln_sa_periods = self._evaluate_gmm(periods)
        shc_list = []
        for itr in range(len(periods)):
            prob_dist_params_im = np.row_stack(
                (sig_ln_sa_periods[:, itr], np.abs(0 * med_sa_periods[:, itr]), med_sa_periods[:, itr]))
            shc_list.append(self._compute_seismic_hazard_integral(lognorm, prob_dist_params_im, im_input)['seismic_hazard_curve'])
        return shc_list

    def select_ground_motion_records(self, n_gm, mean_req, cov_req, spectral_periods, mrp, for_which, **kwargs):
        rng_seed = kwargs.get('rng_seed', None)
        max_scale = kwargs.get('max_scale', 4)
        min_scale = kwargs.get('min_scale', 1 / 3)
        dev_weights = kwargs.get('dev_weights', [1, 2])
        n_loop = kwargs.get('n_loop', 2)
        penalty = kwargs.get('penalty', 0)
        is_scaled = kwargs.get('is_scaled', True)
        not_allowed = kwargs.get('not_allowed', [])
        classify_pulse = kwargs.get('classify_pulse', True)
        sampling_method = kwargs.get('sampling_method', 'mcs')
        deterministic = kwargs.get('deterministic', False)

        ground_motion_records = self._select_ground_motion_records(deterministic, n_gm, mean_req, cov_req, mrp, for_which,
                                                                   spectral_periods,
                                                                   rng_seed, max_scale, min_scale, dev_weights, n_loop,
                                                                   penalty, is_scaled, not_allowed, classify_pulse,
                                                                   sampling_method)

        return ground_motion_records

    # ------------------------------------------------------------------------------------------------------------------

    def get_target_spectrum(self, mrp, for_which, **kwargs):
        uncondition = kwargs.get('uncondition', True)
        spectral_periods = kwargs.get('spectral_periods', np.array([]))
        spectral_periods = np.array(spectral_periods).flatten()
        if spectral_periods.size == 0:
            spectral_periods = np.logspace(np.log10(0.05), np.log10(5), 50, endpoint=True)
        uhs = kwargs.get('uhs', False)

        spectral_periods = Utility.merge_sorted_1d_arrays(spectral_periods, self.evaluate_period(for_which))
        if uhs:
            temp = self.uniform_hazard_spectrum(mrp, spectral_periods=spectral_periods)
            return np.log(temp[:, 1][np.newaxis, :]), np.zeros(shape=(temp.shape[0], temp.shape[0], 1)), temp[:, 0]

        med_sa_spectrum, sig_ln_sa_spectrum = self._evaluate_gmm(spectral_periods)  # (n_s, n_sp), (n_s, n_sp)
        mean_req, cov_req, spectral_periods = self._get_target_spectrum(mrp, for_which,
                                                                        med_sa_spectrum, sig_ln_sa_spectrum,
                                                                        spectral_periods,
                                                                        uncondition)
        return mean_req, cov_req, spectral_periods

    # ------------------------------------------------------------------------------------------------------------------

    def compute_seismic_hazard_integral(self, for_which, **kwargs):
        im_input = kwargs.get('im_input', np.array([]))
        im_input = np.array(im_input).flatten()
        if im_input.size == 0:
            im_input = np.logspace(np.log10(1e-4), np.log10(5), 1000, endpoint=True)

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        med_sa, sig_ln_sa = self.evaluate_gmm(self.evaluate_period(for_which))
        prob_dist_params_im = np.row_stack((sig_ln_sa, np.abs(0 * med_sa), med_sa))
        to_return = self._compute_seismic_hazard_integral(lognorm, prob_dist_params_im, im_input)

        return to_return

    # ------------------------------------------------------------------------------------------------------------------

    def uniform_hazard_spectrum(self, mrp, **kwargs):
        spectral_periods = kwargs.get('spectral_periods', np.array([]))
        spectral_periods = np.array(spectral_periods).flatten()
        if spectral_periods.size == 0:
            spectral_periods = np.logspace(np.log10(0.05), np.log10(5), 50, endpoint=True)

        im_input = kwargs.get('im_input', np.array([]))
        im_input = np.array(im_input).flatten()
        if im_input.size == 0:
            im_input = np.logspace(np.log10(1e-4), np.log10(5), 1000, endpoint=True)

        uhs = []
        shc_list = self.shc_at_periods(spectral_periods, im_input=im_input)
        for itr in range(len(spectral_periods)):
            uhs.append(IM.invert_seismic_hazard(shc_list[itr], mrp))
        return np.column_stack((spectral_periods, uhs))

    def get_moc(self, periods):
        n_p = len(periods)
        moc = np.zeros((n_p, n_p))
        correl_func = self.correl_func

        for j in range(n_p):
            moc[:, j] = correl_func(periods, periods[j])

        return moc

    # ------------------------------------------------------------------------------------------------------------------

    def _gmm(self, scenario_data, periods):
        gmm = self.gmm
        scenario = pygmm.Scenario(**scenario_data)
        return gmm(scenario).interp_spec_accels(periods), gmm(scenario).interp_ln_stds(periods)

    # ------------------------------------------------------------------------------------------------------------------

    def _evaluate_gmm(self, periods):

        if np.isscalar(periods):
            periods = np.array([periods])
        else:
            periods = np.array(periods)

        scenario_data = self.structure.get_site_hazard_info()['scenario_data']

        n_p = len(periods)
        n_s = len(list(scenario_data.values())[0])

        med_sa_periods = np.zeros((n_s, n_p))
        sig_ln_sa_periods = np.zeros((n_s, n_p))

        scenarios = [{key: value for key, value in zip(scenario_data.keys(), values)} for values in zip(*scenario_data.values())]

        for i in range(n_s):
            med_sa_periods[i, :], sig_ln_sa_periods[i, :] = self._gmm(scenarios[i], periods)

        to_return = list()
        to_return.append(med_sa_periods)
        to_return.append(sig_ln_sa_periods)

        return to_return

    # ------------------------------------------------------------------------------------------------------------------

    def get_conditional_spectrum(self, mrp, for_which, med_sa_spectrum, sig_ln_sa_spectrum,
                                 rho, rho_all,
                                 med_sa_star, sig_ln_sa_star):

        # med_sa_spectrum : (n_s, n_sp)
        # sig_ln_sa_spectrum : (n_s, n_sp)
        # rho : (n_s, n_sp)
        # rho_all : (n_sp, n_sp)
        # med_sa_star : (n_s,)
        # sig_ln_sa_star : (n_s,)

        sa_cond_value = self.get_inv_seismic_hazard(mrp, for_which)
        n_s = len(med_sa_star)
        n_sp = rho_all.shape[0]

        # Use the CMS method to estimate target means and covariances
        eps_bar = (np.log(sa_cond_value) - np.log(med_sa_star)) / sig_ln_sa_star  # (n_s,)
        mean_req = np.log(med_sa_spectrum) + sig_ln_sa_spectrum * eps_bar.reshape(n_s, 1) * rho  # (n_s, n_sp)

        sig_1 = np.einsum('ij,il->jli', sig_ln_sa_spectrum, sig_ln_sa_spectrum) * rho_all.reshape(n_sp, n_sp, 1)
        sig_2 = rho * sig_ln_sa_spectrum * sig_ln_sa_star.reshape(n_s, 1)
        sig_2_sig_2_t = np.einsum('ij,il->jli', sig_2, sig_2)
        term_2 = sig_2_sig_2_t * (1 / (sig_ln_sa_star ** 2)).reshape(1, 1, n_s)

        cov_req = sig_1 - term_2  # (n_sp, n_sp, n_s)
        np.einsum('iij->ji', cov_req)[...] = np.abs(np.einsum('iij->ji', cov_req))

        return mean_req, cov_req

    # ------------------------------------------------------------------------------------------------------------------

    def _select_ground_motion_records(self, deterministic, n_gm, mean_req, cov_req, mrp, for_which, spectral_periods,
                                      rng_seed, max_scale, min_scale, dev_weights, n_loop,
                                      penalty, is_scaled, not_allowed, classify_pulse, sampling_method):

        # Extract rec_data
        rec_data_file_path = Utility.get_path(os.path.dirname(__file__), 'Rec_Data', 'rec_data.pickle')
        rec_data = Utility.pickle_load_dict(rec_data_file_path)
        file_name_1 = rec_data['file_name_1']
        file_name_2 = rec_data['file_name_2']
        file_name_vert = rec_data['file_name_vert']
        per_known = rec_data['per_known']
        sa_known = rec_data['sa_known']
        soil_vs_30 = rec_data['soil_vs_30']
        not_allowed_rec_num = rec_data['not_allowed_rec_num']
        not_allowed_ind = np.unique([i - 1 for i in [*not_allowed_rec_num, *not_allowed]])
        pulse_like_records = rec_data['pulse_like_records']

        if deterministic:
            cov_req[...] = 0

        if np.all(cov_req == 0):
            deterministic = True
        else:
            deterministic = False

        if mean_req.shape[0] > 1 and cov_req.shape[2] > 1:
            is_mixture = True
        else:
            is_mixture = False

        if is_mixture:
            scenario_weights = self.get_seismic_hazard_deagg(mrp, for_which)
            threshold = 1.0e-5
            scenario_mask = scenario_weights >= threshold
            scenario_weights = scenario_weights[scenario_mask]
            scenario_weights = scenario_weights / np.sum(scenario_weights)
        else:
            scenario_weights = np.array([1.0])
            scenario_mask = np.array([True])

        if not deterministic:
            # 2 options:
            # 1. mixture CMS
            # 2. non-mixture CMS
            ind = np.where(np.all(cov_req[:, :, 0] == 0, axis=1))[0][0]
            cov_req = np.delete(np.delete(cov_req, ind, axis=0), ind, axis=1)
            mean_req = np.delete(mean_req, ind, axis=1)
            spectral_periods = np.delete(spectral_periods, ind)
            n_sp = len(spectral_periods)
            # Simulate realizations of ground motion spectra
            n_trials = 20
            if is_mixture:
                cov_req = cov_req[:, :, scenario_mask]
                mean_req = mean_req[scenario_mask, :]
                sig_req = np.sqrt(np.einsum('iik->ki', cov_req))
                mean_req_exact, sig_req_exact = Utility.get_exact_mixture_params(mean_req, sig_req, scenario_weights)
                mean_req_exact = mean_req_exact.flatten()
                sig_req_exact = sig_req_exact.flatten()
                prob_dist_params = np.hstack((np.einsum('ijd->idj', mean_req[:, :, np.newaxis]),
                                              np.einsum('ijd->idj', sig_req[:, :, np.newaxis])))
                mixture_list = list()
                for i_sp in range(n_sp):
                    mixture_list.append(mixture(prob_dist_params[:, :, i_sp], weights=scenario_weights, each=norm))
                m = multivariate_nataf(mixture_list, cov_req / np.einsum('ij,ik->jki', sig_req, sig_req))
                gms = np.exp(m.rvs(size=(n_gm, n_trials), random_state=rng_seed, method=sampling_method))
            else:
                mean_req_exact = mean_req[0, :]
                sig_req = np.sqrt(np.diag(cov_req[:, :, 0]))[np.newaxis, :]
                sig_req_exact = sig_req.flatten()
                prob_dist_list = list()
                for i_sp in range(n_sp):
                    prob_dist_list.append(norm(mean_req[0, i_sp], sig_req[0, i_sp]))
                corr_matrix = cov_req[:, :, 0] / np.einsum('i,j->ij', sig_req_exact, sig_req_exact)
                np.fill_diagonal(corr_matrix, 1.0)
                m = multivariate_nataf(prob_dist_list, corr_matrix, corr_matrix_z=corr_matrix)
                gms = np.exp(m.rvs(size=(n_gm, n_trials), random_state=rng_seed, method=sampling_method))
            # gms : (n_gm : 0, n_trials: 1, n_sp: 2)
            # to
            # gms : (n_gm, n_sp, n_trials)
            gms = gms.transpose((0, 2, 1))
            dev_mean_sim = dev_weights[0] * np.sum(np.square(np.mean(np.log(gms), axis=0).T - mean_req_exact), axis=1)
            dev_skew_sim = 0.1 * (dev_weights[0] + dev_weights[1]) * np.sum(np.square(skew(np.log(gms), axis=0).T), axis=1)
            dev_sig_sim = dev_weights[1] * np.sum(np.square(np.std(np.log(gms), axis=0).T - sig_req_exact), axis=1)
            dev_total_sim = dev_mean_sim + dev_sig_sim + dev_skew_sim
            gm = gms[:, :, np.argmin(np.abs(dev_total_sim))]
        else:
            # 3 options:
            # 1. mixture CMS
            # 2. non-mixture CMS
            # 3. UHS (non-mixture by default), still calling the variable mean_req (abuse of notation)
            if is_mixture:
                mean_req = mean_req[scenario_mask, :]
                mean_req_exact, sig_req_exact = Utility.get_exact_mixture_params(mean_req, np.zeros(mean_req.shape), scenario_weights)
                mean_req_exact = mean_req_exact.flatten()
                sig_req_exact = sig_req_exact.flatten()
            else:
                mean_req_exact = mean_req[0, :]
                sig_req_exact = np.zeros(mean_req_exact.shape)
            gm = np.repeat(np.exp(mean_req_exact)[np.newaxis, :], n_gm, axis=0)

        sample_big, scale_fac, mean_req_exact, sig_req_exact, gm = self.setup_psa_spectra_gms(mrp, for_which,
                                                                                              mean_req_exact,
                                                                                              sig_req_exact,
                                                                                              spectral_periods,
                                                                                              gm, sa_known, per_known,
                                                                                              is_scaled)

        # Find best matches to simulated spectra from ground motion database
        rec_id = np.zeros(n_gm).astype(int) - 1  # 1-d array (-1 because 0 is a valid index in python)
        sample_small = np.zeros((n_gm, sample_big.shape[1]))
        final_scale_fac = np.ones(n_gm)  # 1-d array

        if is_scaled:
            discard_mask = np.any([scale_fac > max_scale, scale_fac < min_scale, soil_vs_30 == -1], axis=0)
        else:
            discard_mask = (soil_vs_30 == -1)

        for i in range(n_gm):
            err = np.sum(
                np.square(
                    np.log(
                        np.exp(sample_big) * scale_fac.reshape(len(scale_fac), 1)
                    ) - np.log(gm[i, :]).reshape(1, gm.shape[1])
                ), axis=1
            )

            err[discard_mask] = self.large_err
            err[not_allowed_ind] = self.large_err
            err[rec_id[rec_id >= 0]] = self.large_err

            min_err = np.min(err)
            rec_id[i] = np.argmin(err)
            if min_err >= self.large_err:
                for_which_str = '_'.join(for_which)
                warning_str = f'Warning: Problem with GMS. No good mathces found for for_which = {for_which_str}!'
                print(warning_str)
            if is_scaled:
                final_scale_fac[i] = scale_fac[rec_id[i]]
            sample_small[i, :] = np.log(np.exp(sample_big[rec_id[i], :]) * scale_fac[rec_id[i]])

        # Greedy subset modification procedure
        rec_id, final_scale_fac = greedy(
            n_loop, rec_id, sample_small, sample_big,
            mean_req_exact, sig_req_exact, np.array(dev_weights), scale_fac, final_scale_fac,
            deterministic, penalty, is_scaled, max_scale, min_scale, soil_vs_30, not_allowed_ind, self.large_err
        )

        final_rsn = rec_id + 1
        final_scale_factors = final_scale_fac

        # Output data
        ground_motion_records = dict()
        ground_motion_records['gm_nbr'] = list()
        ground_motion_records['NGA_RSN'] = list()
        ground_motion_records['scale_factor'] = list()
        ground_motion_records['file_name_1'] = list()
        ground_motion_records['file_name_2'] = list()
        ground_motion_records['file_name_vert'] = list()

        if classify_pulse:
            ground_motion_records['is_pulse'] = list()

        for i in range(n_gm):
            ground_motion_records['gm_nbr'].append(i + 1)
            ground_motion_records['NGA_RSN'].append(final_rsn[i])
            ground_motion_records['scale_factor'].append(final_scale_factors[i])
            ground_motion_records['file_name_1'].append(file_name_1[final_rsn[i] - 1].replace('.at2', '.AT2'))
            ground_motion_records['file_name_2'].append(file_name_2[final_rsn[i] - 1].replace('.at2', '.AT2'))
            ground_motion_records['file_name_vert'].append(file_name_vert[final_rsn[i] - 1].replace('.at2', '.AT2'))
            if classify_pulse:
                if final_rsn[i] in pulse_like_records:
                    ground_motion_records['is_pulse'].append(1)
                else:
                    ground_motion_records['is_pulse'].append(0)

        return ground_motion_records

    # ------------------------------------------------------------------------------------------------------------------

    def plot_gm_psa_spectra(self, for_which, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        lc = kwargs.get('lc', 'black')
        lc1 = kwargs.get('lc1', 'red')
        lw = kwargs.get('lw', 1)
        txc = kwargs.get('txc', 'red')
        k = kwargs.get('k', 1.96)
        fig_ax = kwargs.get('fig_ax', None)
        plot_uhs = kwargs.get('plot_uhs', False)
        uhs_lc = kwargs.get('uhs_lc', 'deepskyblue')
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)
            ax = fig.gca()
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        ax.minorticks_on()
        ax.grid(True, which="major", alpha=grid_alpha)
        ax.grid(True, which="minor", alpha=minor_grid_alpha)

        orig_for_which = for_which
        dir_level_to_seek = 'Hazard_Level_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        if len(orig_for_which) > len(for_which):
            gm_num = int(orig_for_which[len(for_which)])
        else:
            gm_num = None

        gm_records, target_spectra, mrp, _ = self.get_gms_results(for_which)
        mean_req, cov_req, spectral_periods = target_spectra

        if mean_req.shape[0] > 1 and len(cov_req.shape) > 2 and cov_req.shape[2] > 1:
            sig_req = np.sqrt(np.einsum('iik->ki', cov_req))
            scenario_weights = self.get_seismic_hazard_deagg(mrp, for_which)
            mean_req_exact, sig_req_exact = Utility.get_exact_mixture_params(mean_req, sig_req, scenario_weights)
            mean_req_exact = mean_req_exact.flatten()
            sig_req_exact = sig_req_exact.flatten()
        else:
            try:
                mean_req_exact = mean_req[0, :]
                sig_req_exact = np.sqrt(np.diag(cov_req[:, :, 0]))
            except IndexError:
                mean_req_exact = mean_req
                sig_req_exact = np.sqrt(np.diag(cov_req[:, :]))

        rec_data_file_path = Utility.get_path(os.path.dirname(__file__), 'Rec_Data', 'rec_data.pickle')
        rec_data = Utility.pickle_load_dict(rec_data_file_path)
        per_known = rec_data['per_known']
        sa_known = rec_data['sa_known']

        # Scale already saved response spectra of eqkes in ensemble. Much faster
        psa = sa_known[np.array(gm_records['NGA_RSN']) - 1, :].T
        scale_fac = np.array(gm_records['scale_factor']).reshape(1, len(gm_records['scale_factor']))
        psa = psa * scale_fac

        for i_nbr in range(len(gm_records['gm_nbr'])):
            ax.plot(per_known, psa[:, i_nbr], color=0.5 * (np.ones(3) - np.array(colors.to_rgb(lc))),
                    alpha=0.5, linewidth=lw / 2)

        if gm_num is not None:
            ax.plot(per_known, psa[:, gm_num - 1], color=lc1, linewidth=lw)

        if plot_uhs:
            uhs = self.uniform_hazard_spectrum(mrp)
            ax.plot(uhs[:, 0], uhs[:, 1], color=uhs_lc, linewidth=lw * 1.5)

        sample_mean = np.mean(np.log(psa), axis=1)
        sample_median = np.exp(sample_mean)
        sample_std = np.std(np.log(psa), axis=1, ddof=1)
        sample_k_lower = np.exp(sample_mean - k * sample_std)
        sample_k_upper = np.exp(sample_mean + k * sample_std)
        ax.plot(per_known, sample_median, color=lc, linestyle='--', linewidth=lw)
        ax.plot(per_known, sample_k_lower, color=lc, linestyle='--', linewidth=lw)
        ax.plot(per_known, sample_k_upper, color=lc, linestyle='--', linewidth=lw)

        median_req_exact = np.exp(mean_req_exact)
        target_median = median_req_exact
        mean_req_exact = np.log(median_req_exact)
        ax.plot(spectral_periods, target_median, color=lc, linestyle='-', linewidth=lw)

        if not np.all(sig_req_exact == 0):
            target_std = sig_req_exact
            target_k_lower = np.exp(mean_req_exact - k * target_std)
            target_k_upper = np.exp(mean_req_exact + k * target_std)
            ax.plot(spectral_periods, target_k_lower, color=lc, linestyle='-', linewidth=lw)
            ax.plot(spectral_periods, target_k_upper, color=lc, linestyle='-', linewidth=lw)

        im_value = self.get_inv_seismic_hazard(mrp, for_which)
        ax.plot(spectral_periods, im_value * np.ones(len(spectral_periods)),
                color=txc, linestyle='-.', linewidth=lw / 2)
        ax.text(spectral_periods[-1], im_value, f'{im_value:.3}', color=txc, horizontalalignment='right',
                verticalalignment='bottom')

        ax.set_xlim([spectral_periods[0], spectral_periods[-1]])

        export_mat_dict = dict()
        export_mat_dict['per_known'] = per_known
        export_mat_dict['psa'] = psa
        if gm_num is not None:
            export_mat_dict['gm_num'] = gm_num
        export_mat_dict['spectral_periods'] = spectral_periods
        export_mat_dict['mean_req'] = mean_req_exact
        export_mat_dict['sig_req'] = sig_req_exact
        export_mat_dict['mrp'] = mrp
        export_mat_dict['im_value'] = im_value

        self.mark_period_on_axis(for_which, ax, **kwargs)
        export_mat_dict['period'] = self.evaluate_period(for_which)

        if save_mat:
            structure = self.structure
            name = structure.name
            for_which_str = '_'.join(for_which)
            file_name = f'plot_gm_psa_spectra_{name}_{for_which_str}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    def plot_unconditional_spectra_pdfs(self, mrp, for_which, **kwargs):
        periods_lim = kwargs.get('periods_lim', (0.05, 5))
        scenario = kwargs.get('scenario', 'uncondition')
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        spectral_periods = kwargs.get('spectral_periods', np.logspace(np.log10(periods_lim[0]),
                                                                      np.log10(periods_lim[-1]),
                                                                      50))

        spectral_periods = np.array(spectral_periods)

        structure = self.structure
        site_hazard_info = structure.get_site_hazard_info()
        scenario_data = site_hazard_info['scenario_data']
        scenario_weights = self.get_seismic_hazard_deagg(mrp, for_which)

        med_sa_spectrum, sig_ln_sa_spectrum = self._evaluate_gmm(spectral_periods)  # (n_s, n_sp), (n_s, n_sp)

        if scenario == 'uncondition':
            mean_ln_sa_spectrum, sig_ln_sa_spectrum = Utility.get_exact_mixture_params(
                np.log(med_sa_spectrum),
                sig_ln_sa_spectrum,
                scenario_weights)  # (1, n_sp), (1, n_sp)
            med_sa_spectrum = np.exp(mean_ln_sa_spectrum).ravel()  # (1, n_sp)
            sig_ln_sa_spectrum = sig_ln_sa_spectrum.ravel()

        if scenario == 'mean':
            numeric_params = [param.name for param in self.gmm.PARAMS if param.__class__.__name__ == 'NumericParameter']
            mean_scenario = {numeric_param: np.sum(scenario_data[numeric_param] * scenario_weights) for numeric_param in numeric_params}
            stacked = np.column_stack([scenario_data[numeric_param] for numeric_param in numeric_params])
            mean_array = np.array(list(mean_scenario.values()))[np.newaxis, :]
            mean_ind = np.argmin(np.sum((stacked - mean_array) ** 2, axis=1))
            mean_scenario = {**mean_scenario, **{key: scenario_data[key][mean_ind] for key in scenario_data.keys() if key not in numeric_params}}
            med_sa_spectrum, sig_ln_sa_spectrum = self._gmm(mean_scenario, spectral_periods)

        if scenario == 'median':
            ind = np.argmin(np.abs(np.cumsum(scenario_weights) - 0.50))
            med_sa_spectrum = med_sa_spectrum[ind, :]
            sig_ln_sa_spectrum = sig_ln_sa_spectrum[ind, :]

        if scenario == 'mode':
            ind = np.argmax(scenario_weights)
            med_sa_spectrum = med_sa_spectrum[ind, :]
            sig_ln_sa_spectrum = sig_ln_sa_spectrum[ind, :]

        if isinstance(scenario, int):
            med_sa_spectrum = med_sa_spectrum[scenario, :]
            sig_ln_sa_spectrum = sig_ln_sa_spectrum[scenario, :]

        if 'spectral_periods' in kwargs.keys():
            kwargs.pop('spectral_periods')
        fig, ax, export_mat_dict = Sa.plot_spectra_pdfs(med_sa_spectrum, sig_ln_sa_spectrum,
                                                        spectral_periods, **kwargs)

        if save_mat:
            name = structure.name
            for_which_str = '_'.join(for_which)
            file_name = f'plot_unconditional_spectra_pdfs_{name}_{for_which_str}_mrp_{mrp}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    def plot_conditional_spectra_pdfs(self, mrp, for_which, **kwargs):
        periods_lim = kwargs.get('periods_lim', (0.05, 5))
        scenario = kwargs.get('scenario', 'uncondition')
        spectral_periods = kwargs.get('spectral_periods', np.logspace(np.log10(periods_lim[0]),
                                                                      np.log10(periods_lim[-1]),
                                                                      50))
        save_mat = kwargs.get('save_mat', False)
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))

        spectral_periods = np.array(spectral_periods)

        structure = self.structure
        site_hazard_info = structure.get_site_hazard_info()
        scenario_data = site_hazard_info['scenario_data']
        scenario_weights = self.get_seismic_hazard_deagg(mrp, for_which)

        ind = 0
        if scenario == 'uncondition':
            mean_req, cov_req, spectral_periods = self.get_target_spectrum(mrp, for_which,
                                                                           spectral_periods=spectral_periods,
                                                                           uncondition=True)
        else:
            mean_req, cov_req, spectral_periods = self.get_target_spectrum(mrp, for_which,
                                                                           spectral_periods=spectral_periods,
                                                                           uncondition=False)

            if scenario == 'mean':
                numeric_params = [param.name for param in self.gmm.PARAMS if param.__class__.__name__ == 'NumericParameter']
                mean_scenario = {numeric_param: np.sum(scenario_data[numeric_param] * scenario_weights) for numeric_param in numeric_params}
                stacked = np.column_stack([scenario_data[numeric_param] for numeric_param in numeric_params])
                mean_array = np.array(list(mean_scenario.values()))[np.newaxis, :]
                mean_ind = np.argmin(np.sum((stacked - mean_array) ** 2, axis=1))
                mean_scenario = {**mean_scenario, **{key: scenario_data[key][mean_ind] for key in scenario_data.keys() if key not in numeric_params}}
                med_sa_spectrum, sig_ln_sa_spectrum = self._gmm(mean_scenario, spectral_periods)
                mean_req, cov_req, spectral_periods = self._get_target_spectrum(mrp, for_which,
                                                                                med_sa_spectrum[np.newaxis, :],
                                                                                sig_ln_sa_spectrum[np.newaxis, :],
                                                                                spectral_periods, False)

            if scenario == 'median':
                ind = np.argmin(np.abs(np.cumsum(scenario_weights) - 0.50))

            if scenario == 'mode':
                ind = np.argmax(scenario_weights)

            if isinstance(scenario, int):
                ind = scenario

        med_sa_spectrum = np.exp(mean_req[ind, :])
        sig_ln_sa_spectrum = np.sqrt(np.diag(cov_req[:, :, ind]))

        if 'spectral_periods' in kwargs.keys():
            kwargs.pop('spectral_periods')
        fig, ax, export_mat_dict = Sa.plot_spectra_pdfs(med_sa_spectrum, sig_ln_sa_spectrum,
                                                        spectral_periods, **kwargs)

        self.mark_period_on_axis(for_which, ax, **kwargs)
        export_mat_dict['period'] = self.evaluate_period(for_which)

        if save_mat:
            name = structure.name
            for_which_str = '_'.join(for_which)
            file_name = f'plot_conditional_spectra_pdfs_{name}_{for_which_str}_mrp_{mrp}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def plot_spectra_pdfs(med_sa_spectrum, sig_ln_sa_spectrum, spectral_periods, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        minor_grid = kwargs.get('minor_grid', False)
        lc = kwargs.get('lc', 'black')
        lw = kwargs.get('lw', 1)
        k = kwargs.get('k', 3)
        n_cont = kwargs.get('n_cont', 100)
        periods_lim = kwargs.get('periods_lim', (0.05, 5))
        plot_periods = kwargs.get('plot_periods', np.logspace(np.log10(periods_lim[0]),
                                                              np.log10(periods_lim[-1]),
                                                              5))
        fig_ax = kwargs.get('fig_ax', None)

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)  # also accepts tight_layout=bool
            ax = fig.add_subplot(projection='3d')
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        if minor_grid:
            ax.minorticks_on()
        else:
            ax.minorticks_off()

        export_mat_dict = dict()

        ax.plot(spectral_periods, med_sa_spectrum, 0, color=lc, linestyle='-.', linewidth=lw)
        ax.plot(spectral_periods, med_sa_spectrum / np.exp(1.96 * sig_ln_sa_spectrum), 0,
                color=lc, linestyle='-.', linewidth=lw / 2)
        ax.plot(spectral_periods, med_sa_spectrum * np.exp(1.96 * sig_ln_sa_spectrum), 0,
                color=lc, linestyle='-.', linewidth=lw / 2)

        export_mat_dict['spectral_periods'] = spectral_periods
        export_mat_dict['med_sa_spectrum'] = med_sa_spectrum
        export_mat_dict['sig_ln_sa_spectrum'] = sig_ln_sa_spectrum
        export_mat_dict['plot_periods'] = plot_periods

        for itr in range(len(plot_periods)):
            ind = np.argmin(np.abs(spectral_periods - plot_periods[itr]))
            med_sa = med_sa_spectrum[ind]
            sig_ln_sa = sig_ln_sa_spectrum[ind]
            # ln_sa = np.logspace(np.log10(1e-4), np.log10(med_sa * np.exp(sig_ln_sa * k)), n_cont)
            ln_sa = np.linspace(1.0e-4, med_sa * np.exp(sig_ln_sa * k), n_cont)
            x = plot_periods[itr] * np.ones(n_cont)
            y = ln_sa
            z = lognorm(sig_ln_sa, 0, med_sa).pdf(y)
            ax.plot(x, y, z, color=lc, linestyle='-', linewidth=lw)
            export_mat_dict[f'ln_sa_period_{itr}'] = y
            export_mat_dict[f'pdf_ln_sa_period_{itr}'] = z

        ax.set_zlim((0, ax.get_zlim()[-1]))
        ax.set_xlim(periods_lim)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------


@jit(nopython=True)
def greedy(n_loop, rec_id, sample_small, sample_big,
           mean_req_exact, sig_req_exact, dev_weights, scale_fac, final_scale_fac,
           deterministic, penalty, is_scaled, max_scale, min_scale, soil_vs_30, not_allowed_ind, large_err):
    n_gm = sample_small.shape[0]
    n_big = sample_big.shape[0]
    for k in range(n_loop):
        for i in range(n_gm):
            min_id = i
            min_dev = large_err
            # sample_small = np.delete(sample_small, i, axis=0)
            # rec_id = np.delete(rec_id, i, axis=0)
            sample_small = np.concatenate((sample_small[:i, :], sample_small[i + 1:, :]), axis=0)
            rec_id = np.concatenate((rec_id[:i], rec_id[i + 1:]), axis=0)

            # Try to add a new spectra to the subset list
            for j in range(n_big):
                temp_sample_small = np.zeros((sample_small.shape[0] + 1, sample_small.shape[1]))
                temp_sample_small[:-1, :] = sample_small
                if is_scaled:
                    # sample_small = np.vstack((sample_small, sample_big[j, :] + np.log(scale_fac[j])))
                    temp_sample_small[-1, :] = sample_big[j, :] + np.log(scale_fac[j])
                else:
                    # sample_small = np.vstack((sample_small, sample_big[j, :]))
                    temp_sample_small[-1, :] = sample_big[j, :]
                sample_small = temp_sample_small

                # Compute deviations from target
                sample_mean = np.sum(sample_small, axis=0) / sample_small.shape[0]
                dev_mean = sample_mean - mean_req_exact
                if not deterministic:
                    sample_std = np.sqrt(np.sum(np.square(sample_small - sample_mean), axis=0) / (sample_small.shape[0] - 1))
                    dev_sig = sample_std - sig_req_exact
                    dev_total = dev_weights[0] * np.sum(np.square(dev_mean)) + dev_weights[1] * np.sum(np.square(dev_sig))
                    # Penalize bad spectra (set penalty to zero if this is not required)
                    if penalty != 0:
                        for m in range(sample_small.shape[0]):
                            dev_total += np.sum(
                                np.exp(sample_small[m, :]) > np.exp(mean_req_exact + 3. * sig_req_exact)
                            ) * penalty
                else:
                    dev_total = dev_weights[0] * np.sum(np.square(dev_mean))

                if is_scaled:
                    if scale_fac[j] > max_scale or scale_fac[j] < min_scale or soil_vs_30[j] == -1 or np.any(not_allowed_ind == j):
                        dev_total += large_err
                else:
                    if soil_vs_30[j] == -1 or np.any(not_allowed_ind == j):
                        dev_total += large_err

                # Should cause improvement and record should not be repeated
                if dev_total < min_dev and not np.any(rec_id == j):
                    min_id = j
                    min_dev = dev_total
                # sample_small = np.delete(sample_small, sample_small.shape[0] - 1, axis=0)
                sample_small = sample_small[:-1, :]

            # Add new element in the right slot
            if is_scaled:
                final_scale_fac[i] = scale_fac[min_id]

            # sample_small = np.vstack((sample_small[:i, :], sample_big[min_id, :] + np.log(scale_fac[min_id]), sample_small[i:, :]))
            # rec_id = np.hstack((rec_id[:i], min_id, rec_id[i:]))
            to_add = sample_big[min_id, :] + np.log(scale_fac[min_id])
            sample_small = np.concatenate((sample_small[:i, :], np.expand_dims(to_add, axis=0), sample_small[i:, :]), axis=0)
            rec_id = np.concatenate((rec_id[:i], np.array([min_id]).astype(rec_id.dtype), rec_id[i:]), axis=0)
    return rec_id, final_scale_fac
